import os

from . import helper
from .ext_exception import ExtException


class OrganizerCode:
    def __init__(self):
        self.code_areas = None
        self.data = None

    @staticmethod
    def is_area(name):
        area_type = name[:17]
        is_area = area_type in ('#Область include_', '#Область includr_')
        if is_area:
            return area_type[9:-1]
        return None

    @classmethod
    def unpack(cls, params):
        src_dir, path, file_name, dest_dir, index, descent = params
        self = cls()
        self.code_areas = {'root': dict(data='')}
        _path = ['']
        _include_path = ['root']
        with open(os.path.join(src_dir, path, file_name), 'r', encoding='utf-8-sig') as file:
            line = file.readline()
            while line:
                try:
                    _line = line.strip()
                    dynamic_directive_index = line.rfind(' //DynamicDirective')
                    if dynamic_directive_index >= 0:
                        if line.find('&') >= 0:
                            line = line[dynamic_directive_index + 1:]
                    if _line and _line[0] == '#':
                        if _line.startswith('#Область'):
                            area_type = cls.is_area(_line)
                            if area_type:
                                area_name_parts = _line[17:].strip().split(' //')
                                key = area_name_parts[0]

                                # if key in self.code_areas:
                                #     raise Exception(
                                #         f'В {path}{file_name} ссылка на один и тот же файл у разных областей {key}')
                                self.code_areas[_include_path[-1]]['data'] += line
                                self.code_areas[key] = dict(data='', area_type=area_type)
                                _include_path.append(key)
                                _path.append(key)
                            else:
                                self.code_areas[_include_path[-1]]['data'] += line
                                _path.append('')

                        elif _line.startswith('#КонецОбласти'):
                            if _path[-1]:  # кончилась include область
                                _include_path.pop()
                            _path.pop()
                            self.code_areas[_include_path[-1]]['data'] += line
                        else:
                            self.code_areas[_include_path[-1]]['data'] += line
                    else:
                        self.code_areas[_include_path[-1]]['data'] += line
                    line = file.readline()
                except Exception as err:
                    raise ExtException(parent=err, detail=f'in file {file_name} line {line}') from err
        return self.code_areas

    @classmethod
    def pack(cls, params):
        src_dir, src_path, src_file_name, dest_dir, dest_path, dest_file_name, index_code_areas, \
        descent, pack_get_descent_filename = params
        data = cls.pack_file(src_dir, src_path, src_file_name, index_code_areas, descent, pack_get_descent_filename)
        helper.txt_write(data, os.path.join(dest_dir, dest_path), dest_file_name)

    @classmethod
    def pack_file(cls, src_dir, path, file_name, index_code_areas, descent, pack_get_descent_filename,
                  dynamic_directive=None):
        normcase_src = os.path.normcase('\\src\\')
        try:
            data = ''
            _src_abs_path = os.path.abspath(os.path.join(src_dir, path))
            if _src_abs_path.startswith(src_dir) or os.path.normcase(path).find(normcase_src) >= 0:
                _src_abs_path, file_name = pack_get_descent_filename(_src_abs_path, file_name, descent)

            with open(os.path.join(_src_abs_path, file_name), 'r', encoding='utf-8') as file:
                line = file.readline()
                while line:
                    _line = line.strip()
                    if _line:
                        if _line.startswith('//DynamicDirective') and dynamic_directive:
                            line = f'{dynamic_directive} {line}'
                    data += line
                    if _line and _line[0] == '#':
                        if cls.is_area(_line):
                            area_name_parts = _line[17:].strip().split(' //')
                            include_path = area_name_parts[0]

                            current_dynamic_directive = area_name_parts[1] if len(area_name_parts) > 1 else dynamic_directive

                            _path, _file_name = cls.parse_include_path(include_path, path, file_name, index_code_areas,
                                                                       descent)
                            _src_abs_path = os.path.abspath(os.path.join(src_dir, _path))
                            if _src_abs_path.startswith(src_dir) or os.path.normcase(_path).find(normcase_src) >= 0:
                                _path, _file_name = pack_get_descent_filename(_src_abs_path, _file_name, descent)
                            data += cls.pack_file(src_dir, _path, _file_name, index_code_areas, descent,
                                                  pack_get_descent_filename, current_dynamic_directive)
                    line = file.readline()
                return data
        except ExtException as err:
            raise ExtException(
                parent=err,
                action=f'{cls.__name__}.pack_file {file_name}') from err
        except Exception as err:
            raise ExtException(
                parent=err,
                action=f'{cls.__name__}.pack_file {file_name}',
                message='Ошибка упаковки файла', detail=f'{os.path.join(path, file_name)}: {err}') from err

    @staticmethod
    def parse_include_path(include_path, path, file_name, index_code_areas, descent=None, *, file_extension='bsl'):
        if index_code_areas and include_path in index_code_areas:
            include_path = index_code_areas[include_path]
        tmp = include_path.split('_')
        size_tmp = len(tmp)
        if size_tmp == 0:
            raise Exception(f'{path} {file_name} в include не указан путь')
        _file_name = f'{tmp[-1]}.{file_extension}'
        _path = '..'  # include не должен лежать внутри папки с исходниками
        if descent is not None:
            _path = os.path.join(_path, '..')
        if size_tmp > 1:
            tmp = ['..' if elem == '' else elem for elem in tmp[:-1]]
            _path = os.path.join(_path, *tmp)
        if _path == '..':
            _path = path
        return _path, _file_name



import os
import shutil
from datetime import datetime

from . import helper
from .ext_exception import ExtException
from .organizer_code import OrganizerCode
from .index import get_dest_path
from .organizer_form_elem import OrganizerFormElem


class OrganizerFile:

    @classmethod
    def unpack(cls, src_dir, dest_dir, *, pool=None, index=None, descent=None):
        begin = datetime.now()
        print(f'{"Организуем код":30}')
        tasks_code_file = []
        tasks_form_elem = []
        helper.clear_dir(dest_dir)
        cls._unpack(src_dir, os.path.abspath(dest_dir), '', tasks_code_file, tasks_form_elem, index,
                    descent)
        helper.run_in_pool(cls.unpack_code_file, tasks_code_file, pool=pool,
                           title=f'{"Раскладываем код по файлам":30}')
        helper.run_in_pool(OrganizerFormElem.unpack, tasks_form_elem, pool=pool,
                           title=f'{"Раскладываем элементы форм":30}')
        print(f'{"Организуем код - готово":30}: {datetime.now() - begin}')

    @classmethod
    def _unpack(cls, src_dir, dest_dir, path, tasks_code_file, tasks_form_elem, index, descent=None):
        entries = os.listdir(os.path.join(src_dir, path))
        for entry in entries:
            try:
                src_entry_path = os.path.join(src_dir, path, entry)

                if os.path.isdir(os.path.normcase(src_entry_path)):
                    new_path = os.path.join(path, entry)
                    cls._unpack(src_dir, dest_dir, new_path, tasks_code_file, tasks_form_elem, index, descent)
                    continue
                if entry.endswith('.bsl'):
                    tasks_code_file.append((src_dir, path, entry, dest_dir, index, descent))
                elif entry.endswith('.elem.json'):
                    tasks_form_elem.append((src_dir, path, entry, dest_dir, index, descent))
                else:
                    src_path = os.path.join(src_dir, path)
                    cls.unpack_file(src_path, entry, dest_dir, path, entry, index, descent)
            except Exception as err:
                raise ExtException(
                    parent=err,
                    action=f'{cls.__name__}._unpack {path}.{entry}'
                ) from err

    @classmethod
    def unpack_file(cls, src_path, src_file_name, dest_dir, dest_path, dest_file_name, index, descent=None):
        try:
            dest_entry_path, dest_file_name = get_dest_path(dest_dir, dest_path, dest_file_name, index, descent)
            dest_full_path = os.path.abspath(os.path.join(dest_dir, dest_entry_path))

            if dest_entry_path:
                helper.makedirs(dest_full_path, exist_ok=True)

            src_full_path = os.path.join(src_path, src_file_name)

            # файлы вне папки исходников не версионируются
            if dest_full_path.startswith(dest_dir) or os.path.normcase(dest_entry_path).find('\\src\\') >= 0:
                descent_full_dest_path, descent_file_name = cls.unpack_get_descent_filename(src_path, src_file_name,
                                                                                            None,
                                                                                            dest_full_path,
                                                                                            dest_file_name,
                                                                                            descent,
                                                                                            cls.equal_binary_file)
            else:
                descent_full_dest_path, descent_file_name = os.path.join(dest_dir, dest_entry_path), dest_file_name

            if descent_file_name:
                shutil.copy(src_full_path, os.path.join(descent_full_dest_path, descent_file_name))
        except Exception as err:
            raise ExtException(parent=err, action=f'{cls.__name__}.unpack_file {src_file_name}') from err

    @classmethod
    def unpack_code_file(cls, params):
        src_dir, path, file_name, dest_dir, index, descent = params
        try:
            code_areas = OrganizerCode.unpack(params)
            for elem in code_areas:
                _file = code_areas[elem]
                if elem == 'root':
                    _file['path'], _file['file_name'] = get_dest_path(dest_dir, path, file_name, index, descent)
                else:
                    if _file['area_type'] == 'includr':
                        continue
                    _file['path'], _file['file_name'] = OrganizerCode.parse_include_path(
                        elem, path, file_name, index.get('Области include') if index else None, descent)
                _file['dest_path'] = os.path.abspath(os.path.join(dest_dir, _file['path']))
                if _file['dest_path'].startswith(dest_dir) or os.path.normcase(_file['path']).find('\\src\\') >= 0:
                    descent_full_dest_path, descent_file_name = cls.unpack_get_descent_filename(
                        None, None, _file['data'], _file['dest_path'], _file['file_name'], descent,
                        cls.equal_code_file)
                else:
                    descent_full_dest_path, descent_file_name = _file['dest_path'], _file['file_name']

                if descent_file_name:
                    helper.txt_write(_file['data'], descent_full_dest_path, descent_file_name)
        except Exception as err:
            raise ExtException(
                parent=err,
                dump={"filename": os.path.join(src_dir, path, file_name)},
                action='unpack_code_file'
            )

    @classmethod
    def equal_binary_file(cls, src_path, src_file_name, src_data, near_descent_path, near_descent_file_name):
        with open(os.path.join(near_descent_path, near_descent_file_name), 'rb') as near_file:
            with open(os.path.join(src_path, src_file_name), 'rb') as src_file:
                near_data = near_file.read()
                src_data = src_file.read()
                if near_data == src_data:  # если файл не менялся, ничего с ним не делаем
                    return True
        return False

    @classmethod
    def equal_code_file(cls, src_path, src_file_name, src_data, near_descent_path, near_descent_file_name):
        near_data = helper.txt_read(near_descent_path, near_descent_file_name)
        if near_data == src_data:  # если файл не менялся, ничего с ним не делаем
            return True
        return False

    @classmethod
    def unpack_get_descent_filename(cls, src_path, src_file_name, src_data, dest_path, dest_file_name, descent,
                                    comparer):
        return dest_path, dest_file_name

    @classmethod
    def pack(cls, src_dir, dest_dir, *, pool=None, index=None, descent=None):
        begin = datetime.now()
        print(f'{"Собираем код":30}')
        try:
            helper.clear_dir(dest_dir)
            tasks_code_file = []
            tasks_form_elem = []
            src_dir = os.path.abspath(src_dir)
            index_code_areas = index.get('Области include') if index else None
            if index:
                cls._pack_index(src_dir, dest_dir, tasks_code_file, tasks_form_elem, index, index_code_areas, [''],
                                descent)
            cls._pack(src_dir, dest_dir, '', tasks_code_file, tasks_form_elem, index, index_code_areas, descent)
            helper.run_in_pool(OrganizerCode.pack, tasks_code_file, pool=pool, title=f'{"Собираем код из файлов":30}')
            helper.run_in_pool(OrganizerFormElem.pack, tasks_form_elem, pool=pool,
                               title=f'{"Собираем элементы форм":30}')
        except Exception as err:
            raise ExtException(parent=err)
        print(f'{"Собираем код - готово":30}: {datetime.now() - begin}')

    @classmethod
    def _pack_index_elem(cls, entry: str, src_dir: str, dest_dir: str, tasks_code_file: list, tasks_form_elem,
                         index_entry: str, index_code_areas: dict, path: list, descent=None):
        normcase_src = os.path.normcase('\\src\\')
        try:
            _src_path = os.path.join(
                '..',
                '' if descent is None else '..',  # в режиме с descent корень находится на уровень выше
                os.path.dirname(index_entry)
            )
            _dest_path = os.path.join(*path)
            if entry.endswith('.bsl'):
                _src_abs_path = os.path.abspath(_src_path)
                if os.path.normcase(_src_path).find(normcase_src) >= 0:
                    func_descent_filename = cls.pack_get_descent_filename
                else:
                    func_descent_filename = OrganizerFile.pack_get_descent_filename
                tasks_code_file.append((
                    src_dir, _src_path, os.path.basename(index_entry),
                    dest_dir, _dest_path, entry, index_code_areas,
                    descent, func_descent_filename))
            elif entry.endswith('.elem.json'):
                tasks_form_elem.append((
                    src_dir, _src_path, os.path.basename(index_entry),
                    dest_dir, _dest_path, entry, index_code_areas, descent,
                    cls.pack_get_descent_filename))
            else:
                _dest_path = os.path.join(dest_dir, *path)
                _src_path = os.path.join(
                    '..',
                    '' if descent is None else '..',  # в режиме с descent корень находится на уровень выше
                    index_entry
                )
                _src_full_path = os.path.join(
                    src_dir,
                    _src_path
                )
                _src_path = os.path.dirname(_src_full_path)
                _src_file_name = os.path.basename(_src_full_path)
                if os.path.normcase(_src_path).find(normcase_src) >= 0:
                    _src_path, _src_file_name = cls.pack_get_descent_filename(_src_path, _src_file_name,
                                                                              descent)
                cls._pack_file(_src_path, _src_file_name, _dest_path, entry, descent)
        except Exception as err:
            raise ExtException(parent=err)

    @classmethod
    def _pack_index(cls, src_dir: str, dest_dir: str, tasks_code_file: list, tasks_form_elem, index: dict,
                    index_code_areas: dict, path: list,
                    descent=None):
        completed_entry = []
        for entry in index:
            try:
                if entry == 'Области include':
                    continue
                if not index[entry]:
                    continue
                if entry.startswith("*"):
                    all_path = index[entry]
                    all_files = os.listdir(all_path)
                    excluded = entry.split('-')[1:]
                    for file_name in all_files:
                        # пропускаем все что были до
                        if file_name in completed_entry:
                            continue
                        if file_name in excluded:
                            continue
                        file_path = os.path.join(all_path, file_name)
                        if not os.path.isfile(os.path.normcase(file_path)):
                            continue
                        cls._pack_index_elem(file_name, src_dir, dest_dir, tasks_code_file, tasks_form_elem, file_path,
                                             index_code_areas, path, descent)
                    continue

                if isinstance(index[entry], dict):
                    path.append(entry)
                    cls._pack_index(src_dir, dest_dir, tasks_code_file, tasks_form_elem, index[entry], index_code_areas,
                                    path, descent)
                    path.pop()
                    pass
                elif isinstance(index[entry], str):
                    completed_entry.append(entry)
                    cls._pack_index_elem(entry, src_dir, dest_dir, tasks_code_file, tasks_form_elem, index[entry],
                                         index_code_areas, path, descent)
                else:
                    raise Exception('Некорректный формат файла индекса')
            except FileNotFoundError:
                pass
            except Exception as err:
                raise ExtException(parent=err)

    @classmethod
    def pack_get_descent_filename(cls, src_path, src_file_name, descent):
        return src_path, src_file_name

    @classmethod
    def list_descent_dir(cls, src_dir, path, descent) -> list[str]:
        return os.listdir(os.path.join(src_dir, path))

    @classmethod
    def _pack(cls, src_dir, dest_dir, path, tasks_code_file, tasks_form_elem, index, index_code_areas, descent=None):
        if path:
            helper.makedirs(os.path.join(dest_dir, path), exist_ok=True)
        entries = cls.list_descent_dir(src_dir, path, descent)
        for entry in entries:
            try:
                src_entry_path = os.path.join(src_dir, path, entry)

                if os.path.isdir(os.path.normcase(src_entry_path)):
                    cls._pack(src_dir, dest_dir, os.path.join(path, entry), tasks_code_file, tasks_form_elem, index,
                              index_code_areas, descent)
                    continue
                _src_path = os.path.join(src_dir, path)
                descent_full_src_path, descent_file_name = cls.pack_get_descent_filename(
                    _src_path, entry, descent)
                if os.path.isfile(os.path.normcase(os.path.join(dest_dir, path, entry))):
                    # если файл уже есть, значит он был переопределен в индексе и делать ничего не надо
                    continue
                if entry.endswith('.bsl'):
                    tasks_code_file.append((
                        src_dir, path, descent_file_name, dest_dir, path, entry, index_code_areas, descent,
                        cls.pack_get_descent_filename))
                elif entry.endswith('.elem.json'):
                    tasks_form_elem.append((
                        src_dir, path, descent_file_name, dest_dir, path, entry, index_code_areas, descent,
                        cls.pack_get_descent_filename))
                else:
                    # shutil.copy(src_entry_path, os.path.join(dest_dir, path, entry))
                    _dest_path = os.path.join(dest_dir, path)
                    cls._pack_file(descent_full_src_path, descent_file_name, _dest_path, entry, descent)
            except Exception as err:
                raise ExtException(parent=err, action=f'{cls.__name__}._pack', detail=f'{entry} {path}')

    @classmethod
    def _pack_file(cls, src_path, src_file_name, dest_path, dest_file_name, descent=None):
        _src_path = os.path.join(src_path, src_file_name)
        _dest_path = os.path.join(dest_path, dest_file_name)
        try:
            shutil.copy(_src_path, _dest_path)
        except FileNotFoundError:
            _dest_dir = os.path.dirname(_dest_path)
            helper.makedirs(_dest_dir, exist_ok=True)
            shutil.copy(_src_path, _dest_path)



import PyInstaller.__main__

PyInstaller.__main__.run([
    'source.py',
    '--onefile',
    '-n', 'v8unpack',
    '--collect-all', 'v8unpack',
    '--hidden-import', 'tqdm',
    '--hidden-import', 'os',
    '--distpath', '.'
])



from .FormElement import FormElement, calc_offset
from v8unpack.helper import calc_offset


class ItemAddition(FormElement):
    @classmethod
    def get_name_node_offset(cls, raw_data):
        return calc_offset([(3, 1), (3, 0)], raw_data)



import os
import sys
import unittest

sys.path.append("../../src/")
from v8unpack.unittest_helper import HelperTestDecode


class TestDecode(HelperTestDecode):
    # processes = 1  # uncomment for debug

    def setUp(self):
        super(TestDecode, self).setUp()
        self.src_dir = os.path.dirname(__file__)
        self.src_file = 'Расширение1.cfe'
        self.init(
            auto_include=True,
            version='80313'
        )
        pass

    def test_01_decode_stage0(self):
        super(TestDecode, self).decode_stage0()

    def test_01_decode_stage1(self):
        super(TestDecode, self).decode_stage1()

    def test_03_decode_stage3(self):
        super(TestDecode, self).decode_stage3()

    def test_04_decode_stage4(self):
        # shutil.rmtree(os.path.join(sys.path[0], 'decodeCodeSubmodule'), ignore_errors=True)
        super(TestDecode, self).decode_stage4()

    def test_05_encode_stage4(self):
        super(TestDecode, self).encode_stage4()

    def test_06_encode_stage3(self):
        super(TestDecode, self).encode_stage3()

    def test_08_encode_stage1(self):
        super(TestDecode, self).encode_stage1()

    def test_09_encode_stage0(self):
        super(TestDecode, self).encode_stage0()


if __name__ == '__main__':
    unittest.main()



from ..MetaDataObject.core.Container import FormContainer


class Document(FormContainer):
    help_file_number = 1
    ext_code = {
        'obj': 0,
        'mgr': 2
    }

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][9][1]




from enum import Enum

from v8unpack import helper
from v8unpack.ext_exception import ExtException


class FormItemTypes(Enum):
    Field = '381ed624-9217-4e63-85db-c4c3cb87daae'
    CheckBox = '35af3d93-d7c7-4a2e-a8eb-bac87a1a3f26'
    RadioBtn = '782e569a-79a7-4a4f-a936-b48d013936ec'
    SelectField = '64483e7f-3833-48e2-8c75-2c31aac49f6e'
    CommandPanel = 'e69bf21d-97b2-4f37-86db-675aea9ec2cb'
    Button = '6ff79819-710e-4145-97cd-1618da79e3e2'
    Image = '151ef23e-6bb2-4681-83d0-35bc2217230c'
    Group = '90db814a-c75f-4b54-bc96-df62e554d67d'
    Table = 'ea83fe3a-ac3c-4cce-8045-3dddf35b28b1'
    TableField = '236a17b3-7f44-46d9-a907-75f9cdc61ab5'
    Panel = '09ccdc77-ea1a-4a6d-ab1c-3435eada2433'
    Label = '0fc7e20d-f241-460c-bdf4-5ad88e5474a5'
    ListField = '19f8b798-314e-4b4e-8121-905b2a7a03f5'
    Separator = '36e52348-5d60-4770-8e89-a16ed50a2006'
    FieldHtml = 'd92a805c-98ae-4750-9158-d9ce7cec2f20'
    Indicator = 'b1db1f86-abbb-4cf0-8852-fe6ae21650c2'
    CalendarBox = 'e3c063d8-ef92-41be-9c89-b70290b5368b'
    TrackBar = '6c06cd5d-8481-4b6f-a90a-7a97a8bb8bef'
    TextDocumentField = '14c4a229-bfc3-42fe-9ce1-2da049fd0109'
    GraphicalSchemaField = '42248403-7748-49da-b782-e4438fd7bff3'
    GeographicalSchemaField = 'ad37194e-555e-4305-b718-5dca84baf145'
    Chart = 'a8b97779-1a4b-4059-b09c-807f86d2a461'
    GanttChart = 'e5fdc112-5c84-4a16-9728-72b85692b6e2'
    PivotChart = 'a26da99e-184a-4823-b0d6-62816d38dc4e'
    Dendrogram = '984981b1-622d-4ebc-94f7-885f0cdfb59a'


class Anchor(Enum):
    top = '0'
    bottom_center = '1'
    left = '2'
    right = '3'


class FormElement:
    ver = 27
    name = 'elements'

    def __init__(self):
        self.anchored = {}

    @classmethod
    def get_class_form_elem(cls, name):
        index = ['Panel']
        if name in index:
            return helper.get_class(f'v8unpack.MetaDataObject.Form.FormElements{cls.ver}.{name}.{name}')
        else:
            return FormElement

    @classmethod
    def set_name(cls, name, raw_data):
        raw_data[-2][1] = helper.str_encode(name)

    @classmethod
    def decode(cls, form, path, elem_raw_data):
        metadata_type_uuid = elem_raw_data[0]
        name = helper.str_decode(elem_raw_data[-2][1])
        try:
            metadata_type = FormItemTypes(metadata_type_uuid)
        except ValueError:
            raise ExtException(
                message='Неизвестный тип элемента формы',
                detail=f'{metadata_type_uuid} {name} - {form.__class__.__name__} {form.header["name"]}',
                action='Form802Element.decode'
            )
        page = elem_raw_data[-3][-5]
        elem_id = elem_raw_data[1]
        elem_data = dict(
            name=name,
            type=metadata_type.name,
            id=elem_id,
            ver=cls.ver
        )
        # cls.decode_anchored(elem_id, form.anchored, path, elem_raw_data[3], 12)
        return form.add_elem(page, path, name, elem_data, elem_raw_data)



    # @classmethod
    # def decode_anchored(cls, anchor_id, anchored, path, elem_raw_data, offset):
    #     def add_anchor():
    #         if len(anchor_data) != 3 or anchor_data[0] != '0':
    #             raise ExtException(message='Неизвестный формат данных привязки', detail=path,
    #                                dump=dict(anchor_data=anchor_data, anchor_id=anchor_id))
    #
    #         elem_id = anchor_data[1]
    #         border = Anchor(anchor_data[2]).name
    #         if elem_id not in anchored:
    #             anchored[elem_id] = []
    #         anchored[elem_id].append(dict(
    #             border=border,
    #             anchor=anchor_id,
    #             anchor_border=Anchor(str(anchor_border)).name,
    #         ))
    #     try:
    #         for anchor_border in range(4):
    #             count_anchor = int(elem_raw_data[offset])
    #             for j in range(count_anchor):
    #                 anchor_data = elem_raw_data[offset + j + 1]
    #                 add_anchor()
    #             elem_raw_data[offset] = '0'
    #             del elem_raw_data[offset + 1: offset + 1 + count_anchor]
    #             offset += 1
    #     except Exception as err:
    #         raise ExtException(parent=err)

    @classmethod
    def encode(cls, form, path, elem_tree, elem_data):
        try:
            raw_data = elem_data['raw']
            if form.auto_include:
                form.last_elem_id += 1
                raw_data[1] = str(form.last_elem_id)

            elem_id = raw_data[1]

            prop = elem_data.get('prop')
            if prop:
                prop_index = form.props_index.get(prop)
                if not prop_index:
                    raise ExtException(message='Отсутствует свойство', detail=prop)
                form.field_data_source.append((elem_id, prop_index))

            # type_index = elem_data.get('type_index')
            # if type_index:
            #     for group in type_index:
            #         for elem in type_index[group]:
            #             form.anchored[group].append([elem[0], elem_id, elem[1]])

            return elem_id, raw_data
        except Exception as err:
            raise ExtException(parent=err)


class FormProps:
    name = 'props'
    name_index = 4


    @classmethod
    def decode(cls, form, elem_raw_data):
        try:
            elem_data = dict(
                name=helper.str_decode(elem_raw_data[cls.name_index]),
                id=elem_raw_data[0][0],
                raw=elem_raw_data
            )
            return elem_data
        except Exception as err:
            raise ExtException(parent=err)

    @classmethod
    def decode_list(cls, form, raw_data, index_element_count=0):
        result = []
        element_count = int(raw_data[index_element_count])
        if not element_count:
            return

        for i in range(element_count):
            elem_raw_data = raw_data[index_element_count + i + 1]
            result.append(cls.decode(form, elem_raw_data))

        raw_data[index_element_count] = 'Дочерние элементы отдельно'
        del raw_data[index_element_count + 1:index_element_count + 1 + element_count]
        return result

    @classmethod
    def encode_list(cls, form, raw_data, path=''):
        try:
            result = []
            index_element_count = 0
            if raw_data[index_element_count] == 'Дочерние элементы отдельно':
                if form.props:
                    for prop in form.props:
                        result.append(cls.encode(form, path, prop))
                    raw_data[index_element_count] = str(len(form.props))
                    raw_data[index_element_count + 1:index_element_count + 1] = result
                else:
                    raw_data[index_element_count] = '0'
        except Exception as err:
            raise ExtException(parent=err)

    @classmethod
    def encode(cls, form, path, data):
        return data['raw']



import os

from . import helper
from .ext_exception import ExtException, KeyNotFound
from .index import get_dest_path
from .organizer_code import OrganizerCode


class OrganizerFormElem:
    def __init__(self):
        self.code_areas = None
        self.data = None

    @classmethod
    def unpack(cls, params):
        src_dir, path, file_name, dest_dir, index, descent = params
        areas = {}
        elements = helper.json_read(os.path.join(src_dir, path), file_name)
        cls._unpack_get_areas(elements['tree'], '', elements['data'], areas)
        areas['root'] = elements
        cls._unpack_write_areas(src_dir, path, file_name, dest_dir, index, descent, areas)

    @staticmethod
    def is_area(name):
        area_type = name[:8]
        is_area = area_type in ('include_', 'includr_')
        if is_area:
            return area_type
        return None

    @classmethod
    def _unpack_get_areas(cls, tree, path, root_data, areas):
        try:
            for elem in tree:
                name: str = elem['name']
                area_type = cls.is_area(name)
                old_path = f'{path}/{name}' if path else name
                if area_type:
                    _path = path[len(path):]
                    new_path = f'{_path}/{name}' if _path else name
                    area_name = name[8:]
                    area = dict(
                        data={
                            # new_path: root_data.pop(old_path)
                        },
                        tree=elem.pop('child', [])
                    )
                    # в пути детей includr не используется.
                    name: str = f'include_{area_name}'
                    old_path = f'{path}/{name}' if path else name
                    cls._pop_area_data(area['tree'], old_path, root_data, area['data'], path)
                    elem['child'] = 'В отдельном файле'
                    if area_type == 'include_':  # includr_ только чтение
                        areas[area_name] = area
                    continue
                child = elem.get('child')
                if child:
                    cls._unpack_get_areas(child, old_path, root_data, areas)

        except Exception as err:
            raise ExtException(parent=err)

    @classmethod
    def _pop_area_data(cls, tree, path, root_data, data, remove_path):
        try:
            size_prefix = len(remove_path)
            if tree:
                for elem in tree:
                    name: str = elem['name']
                    old_path = f'{path}/{name}'
                    new_path = f'{path[size_prefix + 1:] if size_prefix else path}/{name}'
                    try:
                        data[new_path] = root_data.pop(old_path)
                    except Exception:
                        raise KeyNotFound(message='Не найден элемент формы', detail=old_path)
                    child = elem.get('child')
                    if child:
                        cls._pop_area_data(child, old_path, root_data, data, remove_path)

            pages = root_data.pop(f'{path}/-pages-', None)
            if pages is not None:
                new_path = f'{path[size_prefix + 1:] if size_prefix else path}/-pages-'
                data[new_path] = pages
                for name in pages:
                    old_path = f'{path}/{name}'
                    new_path = f'{path[size_prefix + 1:] if size_prefix else path}/{name}'
                    try:
                        data[new_path] = root_data.pop(old_path)
                    except KeyError:
                        pass
                        # raise KeyNotFound(message='Не найден элемент формы', detail=old_path)

        except Exception as err:
            raise ExtException(parent=err)

    @staticmethod
    def _unpack_write_areas(src_dir, path, file_name, dest_dir, index, descent, areas):
        try:
            for elem in areas:
                if elem == 'root':
                    dest_entry_path, dest_file_name = get_dest_path(dest_dir, path, file_name, index, descent)
                else:
                    dest_entry_path, dest_file_name = OrganizerCode.parse_include_path(
                        elem, path, elem, index.get('Области include') if index else None, descent,
                        file_extension=file_name[-9::])
                dest_path = os.path.abspath(os.path.join(dest_dir, dest_entry_path))

                helper.json_write(areas[elem], dest_path, dest_file_name)
        except Exception as err:
            raise ExtException(
                parent=err,
                dump={"filename": os.path.join(src_dir, path, file_name)},
                action='unpack_code_file'
            )

    @classmethod
    def pack(cls, params):
        src_dir, src_path, src_file_name, dest_dir, dest_path, dest_file_name, index_code_areas, \
            descent, pack_get_descent_filename = params
        elements = helper.json_read(os.path.join(src_dir, src_path), src_file_name)
        try:
            cls._pack_get_areas(src_dir, src_path, src_file_name, elements['tree'], '', elements['data'],
                                index_code_areas, descent)
        except Exception as err:
            raise ExtException(parent=err, message='Ошибка сборки элементов формы',
                               detail=f'{os.path.join(src_path, src_file_name)}')

        helper.json_write(elements, os.path.join(dest_dir, dest_path), dest_file_name)

    @staticmethod
    def form_elem_class(version, elem_type):
        elem_class = helper.get_class(f'v8unpack.MetaDataObject.Form.Form{version}Elements.FormElement.FormElement')
        return elem_class.get_class_form_elem(elem_type)

    @classmethod
    def _pack_get_areas(cls, src_dir, src_path, file_name, tree, path, root_data, index_code_areas, descent):
        for elem in tree:
            try:
                name: str = elem['name']
                new_path = f'{path}/{name}' if path else name
                area_type = cls.is_area(name)
                if area_type:
                    area_name = name[8:]
                    _path, _file_name = OrganizerCode.parse_include_path(
                        area_name, src_path, file_name, index_code_areas, descent, file_extension=file_name[-9::])
                    _src_abs_path = os.path.abspath(os.path.join(src_dir, _path))

                    include_elements = helper.json_read(_src_abs_path, _file_name)
                    cls._append_area_data(include_elements['tree'], name, root_data, include_elements['data'],
                                          path, area_type)
                    elem['child'] = include_elements['tree']
                    continue
                child = elem.get('child')
                if child:
                    cls._pack_get_areas(src_dir, src_path, file_name, child, new_path, root_data, index_code_areas,
                                        descent)
            except Exception as err:
                raise ExtException(parent=err, message='Ошибка сборки области формы', detail=elem['name'])

    @classmethod
    def _append_area_data(cls, tree, path, root_data, data, append_path, area_type=None):
        try:
            _path = f'include_{path[8:]}' if area_type == 'includr_' else path
            if tree:
                for elem in tree:
                    name: str = elem['name']
                    old_path = f'{_path}/{name}'
                    new_path = f'{append_path}/{_path}/{name}' if append_path else f'{_path}/{name}'
                    try:
                        root_data[new_path] = data.pop(old_path)
                    except Exception:
                        raise KeyNotFound(message='Не найден элемент формы', detail=old_path)
                    child = elem.get('child')
                    if child:
                        cls._append_area_data(child, old_path, root_data, data, append_path)

            pages = data.pop(f'{path}/-pages-', None)
            if pages is not None:
                new_path = f'{append_path}/{_path}/-pages-' if append_path else f'{_path}/-pages-'
                root_data[new_path] = pages
                for name in pages:
                    old_path = f'{path}/{name}'
                    new_path = f'{append_path}/{_path}/{name}' if append_path else f'{_path}/{name}'
                    try:
                        root_data[new_path] = data.pop(old_path)
                    except KeyError:
                        pass
                        # raise KeyNotFound(message='Не найден элемент формы', detail=old_path)
        except Exception as err:
            raise ExtException(parent=err)

    @staticmethod
    def parse_include_path(include_path, path, file_name, index_code_areas, descent):
        if index_code_areas and include_path in index_code_areas:
            include_path = index_code_areas[include_path]
        tmp = include_path.split('_')
        size_tmp = len(tmp)
        if size_tmp == 0:
            raise Exception(f'{path} {file_name} в include не указан путь')
        _file_name = f'{tmp[-1]}.bsl'
        _path = '..'  # include не должен лежать внутри папки с исходниками
        if descent is not None:
            _path = os.path.join(_path, '..')
        if size_tmp > 1:
            tmp = ['..' if elem == '' else elem for elem in tmp[:-1]]
            _path = os.path.join(_path, *tmp)
        return _path, _file_name



import os
import sys
import unittest

sys.path.append("../../src/")
from v8unpack import helper
from v8unpack.MetaDataObject.versions.Form803 import Form803
from v8unpack.MetaDataObject.versions.Form802 import Form802
from v8unpack.unittest_helper import compare_file, NotEqualLine


class TestFormElem802(unittest.TestCase):
    def setUp(self) -> None:
        self.current_dir = os.path.dirname(__file__)
        self.data_dir = os.path.join(self.current_dir, 'data', 'form82')
        self.temp_dir = os.path.join(self.data_dir, 'temp')
        self.temp_dir2 = os.path.join(self.data_dir, 'temp2')

    def test_convert_to_json(self):
        test_files = os.listdir(self.data_dir)
        for file_name in test_files:
            if os.path.isdir(os.path.join(self.data_dir,file_name)):
                continue
            json_file_name = f'{file_name}.json'
            raw_data = helper.brace_file_read(self.data_dir, file_name)
            helper.json_write(raw_data, self.temp_dir, json_file_name)

    def test_decode_all_form_elem(self):
        test_files = os.listdir(self.data_dir)
        result = ''
        for file_name in test_files:
            if file_name == 'temp':
                continue
            item = file_name.split('.')
            result += self.decode_form_elem(item[0])
        self.assertEqual('', result)

    def test_decode_form_elem(self):
        # file_name = 'СтраницыИПанели'
        # file_name = 'Фильтр'
        # file_name = '2 добавлена страница'
        # file_name = '4 страница 1 добавлена надпись'
        file_name = 'ДоПереносаПоля'
        file_name = 'ПослеПереносаПоля'
        file_name = 'ДоПереноса2'
        file_name = 'ПослеПереноса2'
        result = self.decode_form_elem(file_name)
        self.assertEqual('', result)

    def test_decode_form_elem7(self):
        file_name = 'form7'  # не панель страницы 01 поставлен ручной порядок обхода и изменен порядок 2 и 3 флагов
        result = self.decode_form_elem(file_name)

    def decode_form_elem(self, file_name):
        helper.clear_dir(self.temp_dir)
        json_origin_file_name = f'{file_name}.origin.json'
        json_result_file_name = f'{file_name}.result.json'
        raw_data = helper.brace_file_read(self.data_dir, file_name)
        helper.json_write(raw_data, self.temp_dir, json_origin_file_name)
        form = Form802(options=dict(auto_include=True))
        form.new_dest_dir = self.temp_dir
        form.form = [raw_data]
        form.decode_includes(None, self.temp_dir, '', None)
        form.write_decode_object(self.temp_dir, '', file_name)
        form.encode_nested_includes(self.temp_dir, file_name, self.temp_dir, '')
        helper.json_write(form.form[0], self.temp_dir, json_result_file_name)
        helper.brace_file_write(helper.json_read(self.temp_dir, json_result_file_name), self.temp_dir, file_name)
        problems = ''
        try:
            result = compare_file(
                os.path.join(self.temp_dir, json_origin_file_name),
                os.path.join(self.temp_dir, json_result_file_name),
                problems
            )
        except NotEqualLine as err:
            result = str(err)
        return result



from .FormElement import FormElement, check_count_element
from v8unpack.helper import FuckingBrackets
from v8unpack.ext_exception import ExtException
from v8unpack.helper import calc_offset


class Table(FormElement):

    @classmethod
    def get_name_node_offset(cls, raw_data):
        return calc_offset([(4, 1), (1, 0)], raw_data)

    @classmethod
    def get_prop_link_offset(cls, raw_data):
        return calc_offset([(4, 1), (7, 0)], raw_data)

    @classmethod
    def decode(cls, form, path, raw_data):
        try:
            size = check_count_element([
                (4, 1), (50, 2), (7, 2)
            ], raw_data)
        except Exception as err:
            raise ExtException(parent=err)
        if raw_data[0] == '55' and size != 99:
            raise FuckingBrackets()
        data = super().decode(form, path, raw_data)
        cls.decode_columns(form, path, raw_data, data)
        return data

    @classmethod
    def decode_columns(cls, form, path, raw_data, data):
        try:
            index = calc_offset([
                (4, 1), (50, 2), (7, 0)
            ], raw_data)
            data['child'] = cls.decode_list(form, raw_data, index, f"{path}/{data['name']}" if path else data['name'])
        except Exception as err:
            raise ExtException(parent=err)

    @classmethod
    def encode(cls, form, path, data):
        try:
            try:
                child = data['child']
            except KeyError:
                child = []
            raw_data = super().encode(form, path, data)
            if not child:
                return raw_data
            cls.encode_columns(form, child, raw_data, data, f"{path}")
            return raw_data
        except Exception as err:
            raise ExtException(parent=err)

    @classmethod
    def encode_columns(cls, form, child, raw_data, data, path):
        try:
            index = calc_offset([
                (4, 1), (50, 2), (7, 0)
            ], raw_data)
            cls.encode_list(form, child, raw_data, index, f"{path}/{data['name']}" if path else data['name'])
        except Exception as err:
            raise ExtException(parent=err)


from .FormElement import FormElement, check_count_element
from v8unpack.ext_exception import ExtException
from v8unpack.helper import FuckingBrackets
from v8unpack.helper import calc_offset


class Group(FormElement):
    pass

    @classmethod
    def get_name_node_offset(cls, raw_data):
        return calc_offset([(3, 1), (1, 1), (2, 0)], raw_data)

    @classmethod
    def decode(cls, form, path, raw_data):
        try:
            size = check_count_element([
                (3, 1), (1, 1), (17, 2)
            ], raw_data)
        except Exception as err:
            raise ExtException(parent=err)
        if raw_data[0] == '22' and size < 20:
            raise FuckingBrackets(detail=cls.__name__)

        data = super().decode(form, path, raw_data)
        index = calc_offset([(3, 1), (1, 1), (17, 0)], raw_data)
        new_path = f"{path}/{data['name']}" if path else data['name']
        new_path = new_path.replace('includr_', 'include_')
        data['child'] = cls.decode_list(form, raw_data, index, new_path)
        return data

    @classmethod
    def encode(cls, form, path, data):
        try:
            try:
                child = data['child']
            except KeyError:
                child = []
            raw_data = super().encode(form, path, data)
            if not child:
                return raw_data
            index = calc_offset([(3, 1), (1, 1), (17, 0)], raw_data)
            name_for_child = f"include_{data['name'][8:]}" if data['name'][:8] == 'includr_' else data['name']
            new_path = f"{path}/{name_for_child}" if path else name_for_child
            cls.encode_list(form, child, raw_data, index, new_path)
            return raw_data
        except Exception as err:
            raise ExtException(parent=err)



from .. import helper
from ..ext_exception import ExtException
from ..MetaDataObject.core.Container import Container


class Subsystem(Container):
    help_file_number = 0
    ext_code = {}

    # todo Subsystem не учтено в старых конфакх может быть папка 0 c image и info

    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        super(Subsystem, self).decode_object(src_dir, file_name, dest_dir, dest_path, version, header_data)
        try:
            self.header['info'] = helper.brace_file_read(src_dir, f'{self.header["uuid"]}.1')
        except FileNotFoundError:
            self.header['info'] = None
            return
        pass

    def write_encode_object(self, dest_dir):
        super().write_encode_object(dest_dir)
        try:
            if self.header['info']:
                file_name = f'{self.header["uuid"]}.1'
                helper.brace_file_write(self.header['info'], dest_dir, file_name)
                self.file_list.append(file_name)
        except Exception as err:
            raise ExtException(parent=err,
                               detail=f'{self.__class__.__name__} {self.header["name"]} {self.header["uuid"]}')



from .FormElement import FormElement, FormItemTypes
from v8unpack.helper import calc_offset
from v8unpack import helper
from v8unpack.ext_exception import ExtException


class Panel(FormElement):
    ver = 27

    def __init__(self):
        super().__init__()
        self.elements_tree = []
        self.pages = []
        self.form = None
        self._elements_tree_index = {}
        self.elements_index = {}
        self.anchor_index = [[], [], [], []]
        self.auto_include = False
        self.elements_data = {}
        self.props_index = {}
        self.last_elem_id = 1
        self.field_data_source = []

    @staticmethod
    def calc_id(path, page, name=None):
        id = []
        if path:
            id.append(path)
        if page:
            id.append(page)
        if name:
            id.append(name)
        return '/'.join(id)

    def add_elem(self, page, path, name, elem_data, elem_raw_data):
        def get_page_name():
            if page and self.pages:
                if page != str(int(page)):
                    raise ExtException(message='Не удалось определить номер страницы элемента формы',
                                       detail=f'{path} {name}')
                return self.pages[int(page)]
            return None

        try:
            page_name = get_page_name()
            page_id = self.calc_id(path, page_name, None)
            elem_tree = dict(name=elem_data['name'], type=elem_data['type'])
            if self.auto_include:
                elem_id = self.calc_id(path, page_name, name)
                parent = self.elements_tree[self._elements_tree_index[page_id]]
                parent['child'].append(elem_tree)
            else:
                elem_id = self.calc_id(path, page_name, name)
                elem_tree['page'] = page_name
                self.elements_tree.append(elem_tree)

            self.elements_index[elem_data['id']] = elem_id

            self.elements_data[elem_id] = {
                'id': elem_data['id'],
                'ver': self.ver,
                'page': page_id,
                'raw': elem_raw_data,
            }

            # anchored = self.anchored.get(elem_data['id'])
            # if anchored:
            #     self.elements_data[elem_id]['anchored'] = anchored

            if self.form:
                prop = self.form.props_index.get(elem_data['id'])
                if prop:
                    self.elements_data[elem_id]['prop'] = prop['name']

            return elem_tree, None, elem_id
        except Exception as err:
            raise ExtException(parent=err)

    @classmethod
    def decode(cls, form, path, elem_raw_data):
        try:
            self = cls()
            self.auto_include = form.auto_include
            is_child = 0 if isinstance(elem_raw_data[1], list) else 1
            elem_id = None
            if is_child:
                self.form = form
                self.form.props_index = form.form.props_index
                elem, elem_data, elem_id = super().decode(form, path, elem_raw_data)
                new_path = elem_id.replace('includr_', 'include_')
                self.decode_pages(new_path, elem_raw_data[1 + is_child][1])
                self.decode_elements(new_path, elem_raw_data[-1])
                elem['child'] = self.elements_tree
                form.elements_data.update(self.elements_data)
            else:
                self.form = form
                self.decode_pages(path, elem_raw_data[1 + is_child][1])
                self.decode_elements('', elem_raw_data[2])
            return self.elements_tree, self.elements_data, elem_id
        except Exception as err:
            raise ExtException(parent=err)

    def decode_elements(self, path, raw_data):
        try:
            # result = []
            element_count = int(raw_data[0])
            if not element_count:
                return
            panel_elements = []
            for i in range(element_count):
                elem_raw_data = raw_data[i + 1]
                metadata_type_uuid = elem_raw_data[0]
                try:
                    metadata_type_name = FormItemTypes(metadata_type_uuid)
                except ValueError:
                    raise ExtException(
                        message='Неизвестный тип элемента формы',
                        detail=f'{self.__class__.__name__}: {metadata_type_uuid}'
                    )
                try:
                    handler = self.get_class_form_elem(metadata_type_name.name)
                except Exception as err:
                    raise ExtException(
                        parent=err,
                        message='Проблема с парсером элемента формы',
                        detail=f'{metadata_type_name} - {err}'
                    )

                try:
                    elem_tree, elem_data, elem_id = handler.decode(self, path, elem_raw_data)
                    panel_elements.append(elem_id)
                except helper.FuckingBrackets as err:
                    raise err from err
                except Exception as err:
                    raise ExtException(
                        parent=err,
                        detail=f'{metadata_type_name} - {err}',
                        message='Ошибка разбора элемента формы'
                    )

                # result.append(res)
            raw_data[0] = 'Дочерние элементы отдельно'
            del raw_data[1:1 + element_count]

            if self.auto_include:
                # меняем ид элеменетов на названия полей
                for elem in panel_elements:
                    data = self.elements_data[elem]
                    try:
                        elem_id = data['id']
                    except (KeyError, TypeError):
                        continue
                    self.anchored_elem_id_to_elem_name(data['raw'][-3], path, elem)
                pass

        # записываем привязки по элементам
        # for elem in self.elements_data:
        #     data = self.elements_data[elem]
        #     try:
        #         elem_id = data['id']
        #     except:
        #         continue
        #     if elem_id in self.anchored:
        #         for anchor in self.anchored[elem_id]:
        #             if anchor['anchor'] in self.elements_index:
        #                 # меняем идентификаторы на имена полей
        #                 anchor['anchor'] = self.elements_index[anchor['anchor']]
        #         data['anchored'] = self.anchored[elem_id]

        # return result
        except Exception as err:
            raise ExtException(parent=err)

    def anchored_elem_id_to_elem_name(self, elem_raw_data, path, current_elem):
        try:
            offset = 6
            # к чему привязан этот элемент
            for anchor_border in range(6):
                for j in range(1, 3):
                    elem_id = elem_raw_data[offset][j][1]
                    if int(elem_id) > 0:
                        try:
                            elem_raw_data[offset][j][1] = self.elements_index[elem_id][
                                                          len(path):]  # нужно имя относительно панели
                        except KeyError:  # если такого индекса нет, то это сам элемент - но это не точно
                            elem_raw_data[offset][j][1] = current_elem[len(path):]
                offset += 1
            # кто привязан к этому элементу
            for anchor_border in range(4):
                count_anchor = int(elem_raw_data[offset])
                for j in range(count_anchor):
                    offset += 1
                    anchor_data = elem_raw_data[offset]
                    elem_id = anchor_data[1]
                    if int(elem_id) > 0:
                        try:
                            elem_name = self.elements_index[elem_id][len(path):]
                            anchor_data[1] = elem_name
                        except KeyError:
                            pass  # ошибка, привязан элемент с другой страницы - в интерфейсе такой возможности нет

                # elem_raw_data[offset] = '0'
                # del elem_raw_data[offset + 1: offset + 1 + count_anchor]
                offset += 1
        except Exception as err:
            raise ExtException(parent=err)
        pass

    def decode_pages(self, path, raw_data):
        # def decode_anchored():
        #     self.anchored = {}
        #     type_offset = 2
        #     for border in range(1, 5):
        #         # anchored_elements = {}
        #         # self.anchored.append(anchored_elements)
        #         type_count = int(raw_data[type_offset])
        #         for elem in range(type_count):
        #             _elem_data = raw_data[type_offset + 1 + elem]
        #             self.add_anchor('$Panel', border, _elem_data)
        #             # _elem_id = _elem_data[1]
        #             #
        #             # if _elem_id not in self.anchored:
        #             #     self.anchored[_elem_id] = {}
        #             # if border not in self.anchored[_elem_id]:
        #             #     self.anchored[_elem_id][border] = []
        #             # self.anchored[_elem_id][border] += [['Panel', _elem_data[0], _elem_data[2]]]
        #         raw_data[type_offset] = '0'
        #         del raw_data[type_offset + 1: type_offset + 1 + type_count]
        #         type_offset += 1
        def delete_anchored():
            type_offset = 2
            for border in range(4):
                type_count = int(raw_data[type_offset])
                raw_data[type_offset] = '0'
                del raw_data[type_offset + 1: type_offset + 1 + type_count]
                type_offset += 1

        def decode_info():
            pages_info_offset = pages_offset + 4

            pages_info_count = int(raw_data[pages_info_offset])
            extra_page_info_count = pages_info_count - page_count * 4  # хрен его знает что это, но встречается большее количество записей
            # if page_count * 4 != pages_info_count:
            #     raise NotImplementedError()

            for i in range(page_count):
                offset = i * 4 + 1 + pages_info_offset
                if i == 0:
                    offset = 1 + pages_info_offset
                    page_info = raw_data[offset: offset + 4 + extra_page_info_count]
                else:
                    offset = i * 4 + 1 + pages_info_offset + extra_page_info_count
                    page_info = raw_data[offset: offset + 4]

                elem_id = self.calc_id(path, self.pages[i], None)
                self.elements_data[elem_id]['info'] = page_info

            del raw_data[pages_info_offset + 1: pages_info_offset + 1 + pages_info_count]
            raw_data[pages_info_offset] = 'в отдельном файле'

        pages_offset = self.pages_offset(raw_data)
        try:
            pages_raw_data = raw_data[pages_offset]
            format_version = pages_raw_data[0]
            if format_version != '1':
                print(f'Неизвестный формат страниц. {format_version} != 1')
            page_count = int(pages_raw_data[1])
            self.pages = []
            for i in range(page_count):
                raw_page = pages_raw_data[i + 2]
                page_format_version = raw_page[0]
                # if page_format_version not in ['3', '4']:
                #     print(f'Неизвестный формат страницы. {page_format_version} != 3')
                page_name = helper.str_decode(raw_page[6])
                if self.auto_include:
                    self.elements_tree.append({
                        "name": page_name,
                        "type": "Page",
                        "child": []
                    })
                elem_id = self.calc_id(path, page_name, None)

                self._elements_tree_index[elem_id] = len(self.elements_tree) - 1
                self.pages.append(page_name)

                self.elements_data[elem_id] = {
                    "ver": self.ver,
                    "page_format_version": page_format_version,
                    "raw": raw_page
                }
            del pages_raw_data[1:1 + 2 + page_count]
            self.elements_data[self.calc_id(path, '-pages-', None)] = self.pages

            decode_info()
            if self.auto_include:
                delete_anchored()

        except Exception as err:
            raise ExtException(parent=err)

    @classmethod
    def encode(cls, form, path, elem_tree, elem_data):
        try:
            self = cls()
            self.form = form
            self.auto_include = form.auto_include
            self.props_index = form.props_index
            elem_raw_data = elem_data['raw']
            elem_id = None
            is_child = 0 if isinstance(elem_raw_data[1], list) else 1
            if is_child:
                elem_id, elem_data = super().encode(form, path, elem_tree, elem_data)
                self.last_elem_id = form.last_elem_id

                if self.auto_include:
                    path = self.calc_id(path, elem_tree['name'])
                else:
                    path = self.calc_id(path, elem_tree['page'], elem_tree['name'])
                # path = f"{elem_tree['page']}/{path}/{}" if path else f"{elem_tree['page']}/{elem_tree['name']}"
                self.elements_tree = elem_tree['child']
                self.elements_data = form.elements_data
                elem_raw_data[1 + is_child][1] = self.encode_pages(path, elem_raw_data[1 + is_child][1])
                self.encode_elements(path, elem_raw_data[-1])
                form.last_elem_id = self.last_elem_id
                pass
                # return elem_raw_data
                # elem['child'] = self.elements_tree
                # form.elements_data.update(self.elements_data)
            else:
                self.elements_tree = form.elements_tree
                self.elements_data = form.elements_data
                elem_raw_data[1 + is_child][1] = self.encode_pages(path, elem_raw_data[1 + is_child][1])
                self.encode_elements('', elem_raw_data[2])
            if self.auto_include:  # перезаписываем оглавление привязок
                elem_raw_data[1 + is_child][1] = self.encode_anchor_index(elem_raw_data[1 + is_child][1])
                # super().encode(form, path, raw_data)
            form.field_data_source += self.field_data_source
            return elem_id, elem_raw_data
        except Exception as err:
            raise ExtException(parent=err)

    @staticmethod
    def pages_offset(raw_data):
        try:
            return calc_offset([[2, 1], [1, 1], [1, 1], [1, 1], [1, 1], [1, 1], [4, 0]], raw_data)
        except Exception as err:
            raise ExtException(message='Не смогли найти описание страниц элементов формы')

    def encode_pages(self, path, raw_data):
        def encode_page(_elem_id):
            try:
                # elem_id = '/'.join([_path, _name]) if path else _name
                page_data = self.elements_data[_elem_id]
                pages.append(page_data['raw'])
                return page_data.get('info', [])
            except Exception as err:
                raise ExtException(parent=err)

        _path = path
        _path = path.replace('includr_', 'include_')
        pages_offset = self.pages_offset(raw_data)
        try:
            pages_raw_data = raw_data[pages_offset]
            format_version = pages_raw_data[0]
            if format_version != '1':
                print(f'Неизвестный формат страниц. {format_version} != 1')
            pages = []
            pages_info = []

            if self.auto_include:
                page_count = len(self.elements_tree)
                pages.append(str(page_count))
                for elem in self.elements_tree:
                    pages_info += encode_page(self.calc_id(_path, elem['name']))

            else:
                _pages = self.elements_data[self.calc_id(_path, '-pages-')]
                page_count = len(_pages)
                pages.append(str(page_count))
                for name in _pages:
                    pages_info += encode_page(self.calc_id(_path, name))

            raw_data[pages_offset] = pages_raw_data[:1] + pages + pages_raw_data[1:]

            pages_info_offset = pages_offset + 4
            raw_data = raw_data[:pages_info_offset + 1] + pages_info + raw_data[pages_info_offset + 1:]
            raw_data[pages_info_offset] = str(len(pages_info))
            return raw_data
        except Exception as err:
            raise ExtException(parent=err)

    def encode_anchor_index(self, raw_data):
        try:
            result = []
            for border in range(4):
                anchored = self.anchor_index[border]
                result += [str(len(anchored))]
                if anchored:
                    result += anchored
            raw_data = raw_data[:2] + result + raw_data[6:]
            return raw_data
        except Exception as err:
            raise ExtException(parent=err)

    def encode_element(self, page_id, elem, key, result):
        try:
            try:
                handler = self.get_class_form_elem(elem['type'])
            except Exception as err:
                raise ExtException(
                    parent=err,
                    message='Проблема с парсером элемента формы',
                    detail=f"{elem['type']} - {err}"
                )

            try:
                try:
                    elem_data = self.elements_data[key]
                except KeyError as err:
                    raise ExtException(message='Остутствуют данные элемента формы', detail=key)
                elem_id, res = handler.encode(self, page_id, elem, elem_data)
                self.elements_index[key] = elem_id
            except ExtException as err:
                raise ExtException(
                    parent=err,
                    message='Проблема с парсером элемента формы',
                    detail=f"{elem['type']} - {err}"
                )

            result.append(res)
        except Exception as err:
            raise ExtException(parent=err)

    def encode_elements(self, path, raw_data):

        try:
            _path = path.replace('includr_', 'include_')
            _id = [_path] if _path else []
            self.anchored = {'1': [], '2': [], '3': [], '4': []}
            result = []
            if self.auto_include:
                for i, page in enumerate(self.elements_tree):
                    page_id = '/'.join(_id + [page['name']])
                    if page['child']:
                        for elem in page['child']:
                            key = f"{page_id}/{elem['name']}"
                            self.encode_element(page_id, elem, key, result)
                self.anchored_elem_name_to_elem_id(result, path)
            else:
                for elem in self.elements_tree:
                    page_id = '/'.join(_id)
                    key = []
                    if _path:
                        key.append(_path)
                    if elem['page']:
                        key.append(elem['page'])
                    key.append(elem['name'])
                    key = '/'.join(key)
                    self.encode_element(page_id, elem, key, result)

            raw_data[0] = str(len(result))
            raw_data += result
        except Exception as err:
            raise ExtException(parent=err)

    def anchored_elem_name_to_elem_id(self, result, path):
        def anchor_name_is_num(name):
            try:
                return str(int(name)) == name
            except (TypeError, ValueError):
                return False

        try:
            for elem in result:
                elem_raw_data = elem[-3]
                elem_id = elem[1]
                offset = 6
                _path = path.replace('includr', 'include')
                for anchor_border in range(6):
                    border = 1 if anchor_border > 3 else anchor_border
                    for j in range(1, 3):
                        anchor_name = elem_raw_data[offset][j][1]
                        if anchor_name == '-1':  # ни к чему не привязан
                            continue
                        if anchor_name == '0':  # привязан к панели
                            anchor_border = int(elem_raw_data[offset][j][2])
                            anchor_border = 1 if anchor_border > 3 else anchor_border
                            self.anchor_index[anchor_border].append([elem_raw_data[offset][0], elem_id, str(border)])
                        if anchor_name_is_num(anchor_name):
                            continue
                        else:
                            elem_raw_data[offset][j][1] = self.elements_index[
                                _path + anchor_name]  # нужно имя относительно панели
                    offset += 1

                for anchor_border in range(4):
                    count_anchor = int(elem_raw_data[offset])
                    for j in range(count_anchor):
                        offset += 1
                        anchor_data = elem_raw_data[offset]
                        anchor_name = anchor_data[1]
                        if not anchor_name_is_num(anchor_name):
                            anchor_data[1] = self.elements_index[_path + anchor_name]
                    # elem_raw_data[offset] = '0'
                    # del elem_raw_data[offset + 1: offset + 1 + count_anchor]
                    offset += 1
        except Exception as err:
            raise ExtException(parent=err)
        pass



from ..MetaDataObject.core.Container import Container


class AccountingRegister(Container):
    ext_code = {
        'obj': 6,
        'mgr': 7
    }
    help_file_number = 5

    @classmethod
    def get_decode_header(cls, header):
        obj_version = int(header[0][1][0])
        if obj_version > 21:
            return header[0][1][16][1]
        else:
            return header[0][1][15][1]



import os
import sys
import unittest

sys.path.append("../../src/")
from v8unpack.decoder import Decoder
from v8unpack import helper


class TestDecoder(unittest.TestCase):
    def setUp(self) -> None:
        self.current_dir = os.path.dirname(__file__)
        self.data_dir = os.path.join(self.current_dir, 'data', 'json_decode_src')
        self.temp_dir = os.path.join(self.data_dir, 'temp')

    def test_decode_include(self):
        params = [
            'AccountingRegister',
            ['C:\\project\\v8unpack\\tests_tmp\\erp2\\tmp\\decode_stage_1\\1', '7b248429-2107-4371-b371-a099028cd179',
             'C:\\project\\v8unpack\\tests_tmp\\erp2\\tmp\\decode_stage_3', 'AccountingRegister', {'version': '803'}]]
        tasks = Decoder.decode_include(params)



import os

from .. import helper
from ..MetaObject import MetaObject
from ..ext_exception import ExtException


class MetaDataObject(MetaObject):
    versions = None
    help_file_number = None

    def __init__(self, *, meta_obj_class=None, obj_version=None, options=None):
        super().__init__(meta_obj_class=meta_obj_class, obj_version=obj_version, options=options)
        self.path = ''
        self.new_dest_path = None
        self.new_dest_dir = None
        self.new_dest_file_name = None
        self.parent_id = None

    def get_obj_name(self):
        return self.meta_obj_class.__name__ if self.meta_obj_class else self.__class__.__name__

    @classmethod
    def get_version(cls, header_data, option):
        if cls.versions is None:
            return cls
        return header_data[0][1][0]

    @staticmethod
    def brace_file_read(src_dir, file_name):
        return helper.brace_file_read(src_dir, file_name)

    @classmethod
    def get_handler(cls, header_data, options):
        try:
            # version = helper.get_options_param(options, 'version', '803')
            if cls.versions is None:
                return cls(options=options, obj_version=options['obj_version'])
            obj_version = cls.get_version(header_data, options)
            try:
                return cls.versions[obj_version](options=options, meta_obj_class=cls, obj_version=obj_version)
            except KeyError:
                raise Exception(f'Нет реализации {cls.__name__} для версии "{obj_version}"')
        except Exception as err:
            raise ExtException(message='Не смогли получить класс объекта', detail=f'{cls.__name__} {err}')

    @classmethod
    def decode(cls, src_dir: str, file_name: str, dest_dir: str, dest_path: str, options, *, parent_type=None,
               parent_container_uuid=None):
        try:
            header_data = cls.brace_file_read(src_dir, file_name)
            self = cls.get_handler(header_data, options)
            self.parent_container_uuid = parent_container_uuid
            # self = cls()
            # if parent_type:
            #     self.title = parent_type
            self.decode_header(header_data)
            self.decode_object(src_dir, file_name, dest_dir, dest_path, self.obj_version, header_data)
            tasks = self.decode_includes(src_dir, dest_dir, self.new_dest_path, header_data)
            self.write_decode_object(dest_dir, self.new_dest_path, self.new_dest_file_name)
            return tasks
        except Exception as err:
            problem_file = os.path.join(os.path.basename(src_dir), file_name)
            raise ExtException(
                parent=err,
                message="Ошибка декодирования",
                detail=f'"{cls.__name__}" файл "{problem_file}" ({dest_path})',
                action=f'{cls.__name__}.decode'
            ) from err

    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        self.set_write_decode_mode(dest_dir, dest_path)

    def decode_ids(self):
        return {
            'uuid': self.header.pop('uuid'),
            # 'name': self.header.pop('name')
        }

    def encode_ids(self, data_id):
        self.header['uuid'] = data_id['uuid']

    def write_decode_object(self, dest_dir, dest_path, file_name):
        try:
            dest_full_path = os.path.join(dest_dir, dest_path)

            id_data = self.decode_ids()
            self.header['obj_version'] = self.obj_version
            helper.json_write(id_data, dest_full_path, f'{file_name}.id.json')
            helper.json_write(self.header, dest_full_path, f'{file_name}.json')
            self.write_decode_code(dest_full_path, file_name)
        except Exception as err:
            raise ExtException(parent=err)

    # def set_mode_decode_in_name_folder(self, dest_dir, dest_path):
    #     self.new_dest_path = os.path.join(dest_path, self.header['name'])
    #     self.new_dest_dir = os.path.join(dest_dir, self.new_dest_path)
    #     self.new_dest_file_name = self.get_obj_name()
    #     helper.makedirs(self.new_dest_dir)

    # def set_mode_decode_in_root_folder(self, dest_dir, dest_path):
    #     self.new_dest_path = dest_path
    #     self.new_dest_dir = os.path.join(dest_dir, self.new_dest_path)
    #     self.new_dest_file_name = self.header['name']

    def get_internal_data(self):
        return self.uuid

    @classmethod
    def read_header(cls, src_dir, src_file_name, data_id):
        header = helper.json_read(src_dir, f'{src_file_name}.json')
        # header['name'] = data_id['name']
        header['uuid'] = data_id['uuid']
        return header

    @classmethod
    def encode(cls, src_dir, file_name, dest_dir, parent_id, parent_container_uuid, include_index, options):
        # src_file_name = self.get_encode_file_name(file_name)
        src_file_name = cls.__name__
        try:
            # header_data = header['header']
            # self = cls.get_handler(header_data, options)
            try:
                data_id = helper.json_read(src_dir, f'{src_file_name}.id.json')
                header = helper.json_read(src_dir, f'{src_file_name}.json')
            except FileNotFoundError:
                return None, None
            self = cls.get_handler(header['header'], options)
            self.header = header
            self.encode_ids(data_id)
            self.parent_id = parent_id
            self.container_uuid = self.get_container_uuid(header['header'])
            self.parent_container_uuid = parent_container_uuid


            if not include_index:
                # self = cls(options=options)
                # self.parent_id = parent_id
                # self.parent_container_uuid = parent_container_uuid
                current_obj_id = f"{parent_id}/{cls.__name__}/{file_name}"
                child_tasks = self.encode_includes(src_dir, src_file_name, dest_dir, current_obj_id)
                if child_tasks:
                    object_task = [self.__class__.__name__,
                                   [src_dir, file_name, dest_dir, options, parent_id, self.parent_container_uuid, {}]]
                    return object_task, child_tasks

        except Exception as err:
            raise ExtException(
                parent=err,
                dump=dict(src_dir=src_dir, file_name=file_name),
                action=f'{cls.__name__}.encode') from err

        try:
            self.name = self.header['name']
            self.uuid = data_id['uuid']
            self.obj_version = self.header['obj_version']
            if include_index:
                self.fill_header_includes(include_index)  # todo dynamic index
            self.encode_object(src_dir, src_file_name, dest_dir)
            self.write_encode_object(dest_dir)
            return dict(
                parent_id=parent_id,
                file_list=self.file_list,
                obj_type=self.get_obj_name(),
                obj_name=self.name,
                obj_uuid=self.uuid,
                obj_data=self.get_internal_data()
            ), None
        except Exception as err:
            raise ExtException(
                parent=err,
                dump=dict(uuid=self.header['uuid'], src_dir=src_dir, parent_id=parent_id),
                action=f'{self.__class__.__name__}.encode') from None

    def encode_object(self, src_dir, file_name, dest_dir):
        msg = f'Нет реализации для "{self.__class__.__name__}"'
        raise Exception(msg)

    def write_encode_object(self, dest_dir):
        try:
            file_name = f'{self.header["uuid"]}'
            self.encode_header()
            helper.brace_file_write(self.header['header'], dest_dir, file_name)
            self.file_list.append(file_name)
            self.write_encode_code(dest_dir)
        except Exception as err:
            raise ExtException(parent=err,
                               detail=f'{self.__class__.__name__} {self.header["name"]} {self.header["uuid"]}')

    @staticmethod
    def get_metadata_object_version(path, file_name):
        _path = os.path.join(path, file_name)
        with open(_path, 'r', encoding='utf-8-sig') as file:
            header = file.read(9)
            root = header.split('{')
            if len(root) < 3 or root[1][0] != '1':
                raise ExtException(message='Неизвестная версия объекта метаданных',
                                   detail=file_name)
            root = root[2].split(',')
            return int(root[0])

    def set_write_decode_mode(self, dest_dir, dest_path):
        self.new_dest_path = os.path.join(dest_path, self.header['name'])
        self.new_dest_dir = os.path.join(dest_dir, self.new_dest_path)
        self.new_dest_file_name = self.get_obj_name()
        helper.makedirs(self.new_dest_dir)
        # self.set_mode_decode_in_name_folder(dest_dir, dest_path)

    # def set_write_decode_mode(self, dest_dir, dest_path):
    #     self.set_mode_decode_in_name_folder(dest_dir, dest_path)
    # self.set_mode_decode_in_root_folder(dest_dir, dest_path)

    # def get_encode_file_name(self, file_name):
    #     return self.get_obj_name()

    # def get_encode_file_name(self, file_name):
    #     return file_name



import json
import os

from .helper import remove_descent_from_filename
from .ext_exception import ExtException
from . import helper


def get_from_index(index: dict, path: str, file_name: str):
    try:
        _index = index
        if path:
            _path = os.path.normpath(path)
            _path = path.split(os.sep)
            for _dir in _path:
                _index = _index[_dir]
        try:
            return _index[file_name]
        except KeyError:
            return os.path.join(_index['*'], file_name)
    except TypeError:
        raise ExtException(
            message="Ошибка c уровнями вложенности в Index.json",
            detail=f'{path}/{file_name}',
            action='index.get'
        )


def update_index(src_dir: str, index_file_name: str, dest_dir: str):
    def _update_index(_src_dir, _index, _dest_dir, _path):
        entries = os.listdir(os.path.join(_src_dir, _path))
        for entry in entries:
            entry = remove_descent_from_filename(entry)
            new_path = os.path.join(_path, entry)
            if os.path.isdir(os.path.join(_src_dir, new_path)):
                if entry not in _index:
                    _index[entry] = {}
                _update_index(_src_dir, _index[entry], _dest_dir, new_path)
            else:
                if entry not in _index:
                    _index[entry] = os.path.join(_dest_dir, _path, entry) if _dest_dir else ''
        pass

    try:
        with open(index_file_name, 'r', encoding='utf-8') as f:
            index = json.load(f)
    except FileNotFoundError:
        index = {}

    _update_index(src_dir, index, dest_dir, '')
    with open(index_file_name, 'w+', encoding='utf-8') as f:
        json.dump(index, f, ensure_ascii=False, indent=2)


def get_dest_path(dest_dir: str, path: str, file_name: str, index: dict, descent: int = None):
    try:
        if index:
            try:
                _res = get_from_index(index, path, file_name)
            except KeyError:
                _res = None

            if _res:
                _path = os.path.dirname(_res)
                _file = os.path.basename(_res)
                _path = os.path.join(
                    '..',
                    '' if descent is None else '..',  # в режиме с descent корень находится на уровень выше
                    _path
                )

                try:
                    helper.makedirs(os.path.join(dest_dir, _path), exist_ok=True)
                except FileExistsError:
                    pass
                return _path, _file

        return path, file_name
    except Exception as err:
        raise ExtException(
            parent=err,
            message='Ошибка получения пути из index.json',
            detail=f'{path}\{file_name}',
            action='CodeOrganizer.get_dest_path',
        ) from err



from .FormElement import FormProps
from .Panel import Panel
from v8unpack import helper
from v8unpack.ext_exception import ExtException


class FormElements27:
    FormProps = FormProps
    Panel = Panel

    def __init__(self, form):
        self.form = form
        self.props_index = {}
        self.last_elem_id = 1
        self.field_data_source = []

    @property
    def options(self):
        return self.form.options

    @property
    def auto_include(self):
        if self.form.options:
            return self.form.options.get('auto_include')
        return False

    @property
    def elements_data(self):
        return self.form.elements_data

    @property
    def elements_tree(self):
        return self.form.elements_tree

    def decode(self, src_dir, dest_dir, dest_path, raw_data):
        try:
            self.form.props = self.FormProps.decode_list(self.form, raw_data[2][2])
            self.form.elements_tree, self.form.elements_data = self.decode_elements(raw_data)
        except Exception as err:
            raise ExtException(parent=err)

    def decode_elements(self, form_data):
        try:
            meta_type = form_data[1][2][0]
            if meta_type != '09ccdc77-ea1a-4a6d-ab1c-3435eada2433':
                raise ExtException(message=f"Неизвестный формат элементов формы",
                                   detail=f"Новый тип элементов формы {meta_type}, "
                                          f"просьба передать файл формы {self.form.header.get('name')} разработчикам")
            self.create_prop_index_by_elem_id(form_data[2][3])

            elements_tree, elements_data, elements_id = self.Panel.decode(self, '', form_data[1][2])
            return elements_tree, elements_data
        except Exception as err:
            raise ExtException(parent=err)

    def create_prop_index_by_elem_id(self, raw_data):
        try:
            self.props_index = {}
            _props = {}
            if not self.form.props:
                return
            for prop in self.form.props:
                _props[prop['id']] = prop

            element_count = int(raw_data[0])
            if not element_count:
                return

            for i in range(element_count):
                elem_raw_data = raw_data[i + 1]
                elem_id = elem_raw_data[0]
                # if elem_raw_data[1][0] == '1':
                prop_id = elem_raw_data[1][1][0]
                try:
                    self.props_index[elem_id] = {'name': _props[prop_id]['name'], 'index': elem_raw_data[1]}
                except KeyError:
                    pass
                # else:
                #     raise NotImplementedError('prop index  > 1')
        except Exception as err:
            raise ExtException(parent=err)

    def create_prop_index_by_name(self):
        try:
            self.props_index = {}
            if self.form.props:
                for prop in self.form.props:
                    self.props_index[prop['name']] = prop['id']
        except Exception as err:
            raise ExtException(parent=err)

    def fill_datasource(self, raw_data):
        raw_data.append(str(len(self.field_data_source)))
        for elem_id, prop_id in self.field_data_source:
            raw_data.append(
                [str(elem_id), ['1', [str(prop_id)]]]
            )

    def encode(self, src_dir, file_name, version, raw_data, path=''):
        try:
            # index_element_count = 0
            # if raw_data[index_element_count] == 'Дочерние элементы отдельно':
            # elements = helper.json_read(src_dir, f'{file_name}.elements802.json')
            elements = helper.json_read(src_dir, f'{file_name}.elem.json')
            self.form.props = elements['props']
            self.form.elements_data = elements['data']
            self.form.elements_tree = elements['tree']
            self.FormProps.encode_list(self.form, raw_data[2][2])
            self.create_prop_index_by_name()
            self.Panel.encode(self, '', None, dict(raw=raw_data[1][2]))
            if self.auto_include:
                raw_data[2][3] = []
                self.fill_datasource(raw_data[2][3])
            return raw_data
        except Exception as err:
            raise ExtException(parent=err)



import os
import sys
import unittest

sys.path.append("../../src/")
# from v8unpack.MetaDataObject.versions.Template803 import Meta
from v8unpack.decoder import Decoder
from v8unpack.helper import clear_dir


class TestTemplate(unittest.TestCase):
    def setUp(self) -> None:
        self.current_dir = os.path.dirname(__file__)
        self.data_dir = os.path.join(self.current_dir, 'data')
        self.temp_dir = os.path.join(self.current_dir, 'temp')
        clear_dir(self.temp_dir)

    def test_decode(self):
        Decoder.decode_include(['Template', [
            self.data_dir, '02f0dbd6-4bb9-4fdb-9c02-116da924943f', self.temp_dir, '', '803']])





from enum import Enum

from v8unpack import helper
from v8unpack.ext_exception import ExtException
from v8unpack.helper import calc_offset


def check_count_element(counters, raw_data):
    # counters - позиции указывающие на счетчики, если не 0 то за ним идет столько записей размера size
    index = 0
    var_len = 0
    for counter_index, size in counters:
        index += counter_index
        if size:
            value = int(raw_data[index])
            index += value * size
            var_len += value * size
    return len(raw_data) - var_len


class FormItemTypes(Enum):
    Field = '77ffcc29-7f2d-4223-b22f-19666e7250ba'
    Button = 'a9f3b1ac-f51b-431e-b102-55a69acdecad'
    Decoration = '3d3cb80c-508b-41fa-8a18-680cdf5f1712'
    Group = 'cd5394d0-7dda-4b56-8927-93ccbe967a01'
    Table = '143c00f7-a42d-4cd7-9189-88e4467dc768'
    ItemAddition = 'c5259a1d-518a-4afd-b98d-0176027e4feb'


class FormElement:

    @classmethod
    def get_name_node_offset(cls, raw_data):
        return calc_offset([(3, 1), (1, 0)], raw_data)

    @classmethod
    def get_prop_link_offset(cls, raw_data):
        return None

    @classmethod
    def set_name(cls, name, raw_data):
        name_offset = cls.get_name_node_offset(raw_data)
        raw_data[name_offset] = helper.str_encode(name)

    @classmethod
    def get_command_link_offset(cls, raw_data):
        return None

    @classmethod
    def decode(cls, form, path, raw_data):
        try:
            name_offset = cls.get_name_node_offset(raw_data)
            name = raw_data[name_offset]
            prop_link_offset = cls.get_prop_link_offset(raw_data)
            prop = ''
            if prop_link_offset is not None:
                prop_link = raw_data[prop_link_offset]
                if prop_link and int(prop_link[0]):
                    prop = []
                    prop_src = form.props_index
                    for i in range(int(prop_link[0])):
                        prop_id = prop_link[i + 1][0]
                        try:
                            prop_name = prop_src[prop_id]['name']
                            prop_src = prop_src[prop_id]['child']
                            prop.append(prop_name)
                        except KeyError:
                            prop = None
                            break
                    if prop:
                        prop = '.'.join(prop)

            command_link_offset = cls.get_command_link_offset(raw_data)
            command = []
            if command_link_offset is not None:
                command_link = raw_data[command_link_offset]
                if command_link and int(command_link[0]):
                    command = form.commands_index.get(command_link[0])

            if not isinstance(name, str) or not name:
                raise ExtException(message='form elem name not string')
            data = dict(raw=raw_data, ver=4)
            if prop:
                data['ПутьКДанным'] = prop
            if command:
                data['ИмяКоманды'] = command
            name = helper.str_decode(name)
            key = f'{path}/{name}' if path else name
            form.elements_data[key] = data
            return dict(
                name=name,
                type=cls.__name__,
            )
        except Exception as err:
            raise ExtException(parent=err)

    @classmethod
    def decode_list(cls, form, raw_data, index_element_count, path=''):
        try:
            result = []
            element_count = int(raw_data[index_element_count])
            if not element_count:
                return

            for i in range(element_count):
                metadata_type_uuid = raw_data[index_element_count + i * 2 + 1]
                elem_raw_data = raw_data[index_element_count + i * 2 + 2]
                try:
                    metadata_type = FormItemTypes(metadata_type_uuid)
                except ValueError:
                    raise ExtException(
                        message='Неизвестный тип элемента формы',
                        detail=f'{form.__class__.__name__} {form.header.get("name")} : {metadata_type_uuid}'
                    )
                elem_data = cls.decode_elem(metadata_type.name, form, path, elem_raw_data)
                result.append(elem_data)

            raw_data[index_element_count] = 'Дочерние элементы отдельно'
            del raw_data[index_element_count + 1:index_element_count + 1 + element_count * 2]
            return result
        except helper.FuckingBrackets as err:
            raise err
        except Exception as err:
            raise ExtException(parent=err)

    @classmethod
    def decode_elem(cls, metadata_type_name, form, path, elem_raw_data):
        try:
            handler = cls.get_class_form_elem(metadata_type_name)
        except Exception as err:
            raise ExtException(
                parent=err,
                message='Проблема с парсером элемента формы',
                detail=f'{metadata_type_name} - {err}'
            )
        try:
            return handler.decode(form, path, elem_raw_data)
        except helper.FuckingBrackets as err:
            raise err from err
        except Exception as err:
            raise ExtException(
                parent=err,
                detail=f'{metadata_type_name} - {err}',
                message='Ошибка разбора элемента формы'
            )

    @staticmethod
    def get_class_form_elem(name):
        return helper.get_class(f'v8unpack.MetaDataObject.Form.FormElements4.{name}.{name}')

    @classmethod
    def encode_list(cls, form, items, raw_data, index_element_count, path=''):
        try:
            result = []
            for item in items:
                try:
                    metadata_type = FormItemTypes[item['type']]
                except ValueError:
                    raise ExtException(
                        message='Неизвестный тип элемента формы',
                        detail=f'{form.__class__.__name__} {form.header["name"]} : {item["type"]}'
                    )
                try:
                    handler = cls.get_class_form_elem(metadata_type.name)
                except Exception as err:
                    raise ExtException(
                        message='Проблема с парсером элемента формы',
                        detail=f'{metadata_type.name} - {err}'
                    )
                elem_data = handler.encode(form, path, item)
                result.append(metadata_type.value)
                result.append(elem_data)
            raw_data[index_element_count] = str(len(items))
            raw_data[index_element_count + 1:index_element_count + 1] = result

            return result
        except Exception as err:
            raise ExtException(parent=err)

    @classmethod
    def encode(cls, form, path, data):
        try:
            key = f"{path}/{data['name']}" if path else data['name']
            try:
                elem_data = form.elements_data[key]
            except KeyError as err:
                raise ExtException(message='Остутствуют данные элемента формы', detail=err)
            includr_index = key.find('includr_')
            if includr_index >= 0:
                handler = cls.get_class_form_elem(data['type'])
                name_offset = handler.get_name_node_offset(elem_data['raw'])
                elem_data['raw'][name_offset] = helper.str_encode(data['name'])
                # return elem_data['raw']
            include_index = key.find('include_')
            if include_index >= 0 or includr_index >= 0:
                prop = elem_data.get('ПутьКДанным')
                if prop:
                    raw_data = elem_data['raw']
                    prop_link_offset = cls.get_prop_link_offset(raw_data)
                    if prop_link_offset is not None:
                        prop_link = raw_data[prop_link_offset]
                        if prop_link:
                            prop_index = form.props_index.get(prop)
                            if prop_index:
                                count = int(prop_link[0])
                                if count == 1:
                                    prop_link[1] = [prop_index]
                                    pass
                                elif count == 2:
                                    prop_link[1] = [prop_index[0]]
                                    prop_link[2] = [prop_index[1]]
                                    pass
                                elif count == 0:
                                    pass
                                else:
                                    raise NotImplementedError()
                            else:
                                raise ExtException(
                                    message='Нет свойства',
                                    detail=f'{prop} на форме {form.form.name}',
                                    action=f'{cls.__name__} FormElement.encode'
                                )
                    pass

                command = elem_data.get('ИмяКоманды')
                if command:
                    raw_data = elem_data['raw']
                    command_link_offset = cls.get_command_link_offset(raw_data)
                    if command_link_offset:
                        command_link = raw_data[command_link_offset]
                        if command_link:
                            command_index = form.commands_index.get(command)
                            if command_index:
                                raw_data[command_link_offset] = command_index
                            else:
                                raise ExtException(
                                    message='Нет команды',
                                    detail=f'{command} в форме {form.form.name}',
                                    action=f'{cls.__name__} FormElement.encode'
                                )
                    pass
                return elem_data['raw']

            return elem_data['raw']
        except Exception as err:
            raise ExtException(parent=err)


class _FormRoot:
    element_node_offset = [[2, 1], [3, 1], [4, 1]]
    name = ''
    index = 0
    index_name = 0

    @classmethod
    def decode(cls, form, raw_data):
        return dict(
            name=helper.str_decode(raw_data[cls.index_name]),
            raw=raw_data
        )

    @classmethod
    def decode_list(cls, form, raw_data):
        try:
            if len(raw_data) <= cls.index:
                return
            items = raw_data[cls.index]
            index_element_count = 1
            element_count = int(items[index_element_count])
            if not element_count:
                return

            result = []
            for i in range(element_count):
                result.append(cls.decode(form, items[i + index_element_count + 1]))
            items[index_element_count] = 'Дочерние элементы отдельно'
            del items[index_element_count + 1:index_element_count + 1 + element_count]
            return result
        except Exception as err:
            pass

    @classmethod
    def encode_list(cls, form, src_dir, file_name, version, raw_data):
        if len(raw_data) <= cls.index:
            return
        raw_data = raw_data[cls.index]
        result = []
        index_element_count = 1
        if raw_data[index_element_count] == 'Дочерние элементы отдельно':
            items = getattr(form.form, cls.name)  # helper.json_read(src_dir, f'{file_name}.{cls.name}{version}.json')
            if items:
                for item in items:
                    result.append(cls.encode(form, item))
                raw_data[index_element_count] = str(len(items))
                raw_data[index_element_count + 1:index_element_count + 1] = result
            else:
                raw_data[index_element_count] = '0'
            return items

    @classmethod
    def encode(cls, form, data):
        return data['raw']


class FormParams(_FormRoot):
    name = 'params'
    index = 4
    index_name = 1


class FormProps(_FormRoot):
    name = 'props'
    index = 3
    index_name = 3
    child_offset = 13

    @classmethod
    def decode(cls, form, raw_data):
        try:
            result = dict(
                name=helper.str_decode(raw_data[cls.index_name]),
                id=raw_data[1][0],
                raw=raw_data
            )
            child_count = int(raw_data[cls.child_offset])
            pattern = raw_data[5]
            if pattern and len(pattern) > 1 and pattern[1][0] == '"#"' and form.form.parent_container_uuid == \
                    pattern[1][1]:
                pattern[1][1] = "Родитель"

            if child_count:
                result['child'] = []
                for i in range(child_count):
                    child = raw_data[cls.child_offset + i + 1]
                    result['child'].append(cls.decode_child(child))
                raw_data[cls.child_offset] = "отдельно"
                del raw_data[cls.child_offset + 1:cls.child_offset + 1 + child_count]

            return result
        except Exception as err:
            raise ExtException(parent=err)

    @classmethod
    def decode_child(cls, raw_data):
        return dict(
            name=helper.str_decode(raw_data[cls.index_name]),
            id=raw_data[1],
            raw=raw_data
        )

    @classmethod
    def encode(cls, form, data):
        raw_data = data['raw']
        pattern = raw_data[5]
        if pattern and len(pattern) > 1 and pattern[1][0] == '"#"' and \
                pattern[1][1] == 'Родитель' and form.form.parent_container_uuid:
            pattern[1][1] = form.form.parent_container_uuid
        if data.get('child'):
            child_count = len(data['child'])
            raw_data[cls.child_offset] = str(child_count)
            childs = []
            for i in range(child_count):
                child = data['child'][i]
                childs.append(child['raw'])
            raw_data[cls.child_offset + 1: cls.child_offset + 1] = childs
        return raw_data


class FormCommands(_FormRoot):
    name = 'commands'
    index = 5
    index_name = 2

    @classmethod
    def decode(cls, form, raw_data):
        result = dict(
            name=helper.str_decode(raw_data[cls.index_name]),
            id=raw_data[1][0],
            raw=raw_data
        )
        return result



import os
import sys
import unittest

sys.path.append("../../src/")
from v8unpack.MetaObject.Configuration802 import Configuration802
from v8unpack.decoder import Decoder
from v8unpack.helper import clear_dir


class TestTemplate(unittest.TestCase):
    def setUp(self) -> None:
        self.current_dir = os.path.dirname(__file__)
        self.data_dir = os.path.join(self.current_dir, 'data')
        self.temp_dir = os.path.join(self.current_dir, 'temp')
        self.temp_dir2 = os.path.join(self.current_dir, 'temp2')

    def test_decode(self):
        clear_dir(self.temp_dir)
        clear_dir(self.temp_dir2)
        Configuration802.decode(self.data_dir, self.temp_dir)
        Configuration802.encode(self.temp_dir, self.temp_dir2)





from ..MetaDataObject.core.Container import FormContainer
import os

from .. import helper
from ..MetaObject import MetaObject
from ..ext_exception import ExtException


class DataProcessor(FormContainer):
    ext_code = {
        'mgr': '2',  # модуль менеджера
        'obj': '0',  # модуль объекта
    }
    help_file_number = 1

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][3][1]

    def decode_ids(self):
        data_id = super().decode_ids()

        if self.obj_version.startswith('803'):
            manager_data = self.header['header'][0][1]
            data_id['manager_uuid1'] = manager_data[1]
            data_id['manager_uuid2'] = manager_data[2]
            data_id['manager_uuid3'] = manager_data[7]
            data_id['manager_uuid4'] = manager_data[8]
            manager_data[1] = 'manager_uuid1 в файле id'
            manager_data[2] = 'manager_uuid2 в файле id'
            manager_data[7] = 'manager_uuid3 в файле id'
            manager_data[8] = 'manager_uuid4 в файле id'

        return data_id

    def encode_ids(self, data_id):
        super().encode_ids(data_id)
        if self.obj_version.startswith('803'):
            manager_data = self.header['header'][0][1]
            manager_data[1] = data_id['manager_uuid1']
            manager_data[2] = data_id['manager_uuid2']
            manager_data[7] = data_id['manager_uuid3']
            manager_data[8] = data_id['manager_uuid4']

    def encode_header(self):
        super().encode_header()
        self.set_product_comment(self.options.get('product_version'))



from .FormElement import FormProps
from .Panel import Panel
from ..FormElements27.FormElements27 import FormElements27


class FormElements26(FormElements27):
    FormProps = FormProps
    Panel = Panel



from v8unpack.MetaDataObject.Form import Form0


class ChartOfCharacteristicTypeForm(Form0):
    # @classmethod
    # def get_form_root(cls, header_data):
    #     obj_version = header_data[0][1][0]
    #     if obj_version == '9':
    #         return header_data
    #     else:
    #         return header_data[0]

    pass



import os
import shutil
import sys
import unittest

sys.path.append("../../src/")
from v8unpack.unittest_helper import HelperTestDecode


class TestDecode(HelperTestDecode):
    processes = 1  # uncomment for debug

    def setUp(self):
        super(TestDecode, self).setUp()
        self.src_dir = os.path.dirname(__file__)
        self.dest_dir = os.path.join(self.src_dir, 'src')
        self.src_file = 'ВнешняяОбработка803.epf'

        self.init(
            index_file_name='index.json',
            version='803'
        )
        pass

    def test_01_decode_stage0(self):
        super(TestDecode, self).decode_stage0()

    def test_01_decode_stage1(self):
        super(TestDecode, self).decode_stage1()

    def test_03_decode_stage3(self):
        super(TestDecode, self).decode_stage3()
        self.assert_external_data_processor_decode_stage3()

    def test_04_decode_stage4(self):
        # shutil.rmtree(os.path.join(self.test_dir, 'decodeCodeSubmodule'), ignore_errors=True)
        super(TestDecode, self).decode_stage4()

    def test_05_encode_stage4(self):
        super(TestDecode, self).encode_stage4()

    def test_06_encode_stage3(self):
         super(TestDecode, self).encode_stage3()

    def test_08_encode_stage1(self):
        super(TestDecode, self).encode_stage1()

    def test_09_encode_stage0(self):
        super(TestDecode, self).encode_stage0()

    @unittest.skip
    def test_extract_all(self):
        super(TestDecode, self).extract_all('./product.json', 'ВнешняяОбработка803.epf')

    @unittest.skip
    def test_build_all(self):
        super(TestDecode, self).build_all('./product.json', 'ВнешняяОбработка803.epf')

    @unittest.skip
    def test_extract(self):
        from v8unpack.v8unpack import extract, build
        extract('ВнешняяОбработка803.epf', 'src')
        build('src', 'ВнешняяОбработка.build.epf')

    @unittest.skip
    def test_create_index(self):
        from v8unpack.index import update_index
        update_index(self.decode_dir_stage4, os.path.join(sys.path[0], 'tmp', 'index.json'),
                     'decodeCodeSubmodule')



if __name__ == '__main__':
    unittest.main()



from .FormElement import FormProps, FormParams, FormCommands, FormElement
from v8unpack.MetaDataObject import helper
from v8unpack.ext_exception import ExtException
from v8unpack.helper import calc_offset


class FormElements4:

    def __init__(self, form):
        self.form = form
        self.props_index = None
        self.commands_index = None

    @property
    def elements_data(self):
        return self.form.elements_data

    @property
    def elements_tree(self):
        return self.form.elements_tree

    def decode(self, src_dir, dest_dir, dest_path, raw_data):
        try:
            self.form.props = FormProps.decode_list(self, raw_data)
            self.form.commands = FormCommands.decode_list(self, raw_data)
            self.decode_elements(raw_data)
            self.form.params = FormParams.decode_list(self, raw_data)
        except Exception as err:
            raise ExtException(parent=err)

    def create_prop_index_by_id(self):
        self.props_index = {}
        if self.form.props:
            for prop in self.form.props:
                self.props_index[prop['id']] = {'name': prop['name'], 'child': {}}
                childs = prop.get('child', [])
                for child in childs:
                    self.props_index[prop['id']]['child'][child['id']] = {'name': child['name'], 'child': {}}

    def create_commands_index_by_id(self):
        self.commands_index = {}
        if self.form.commands:
            for command in self.form.commands:
                self.commands_index[command['id']] = command['name']

    def create_prop_index_by_name(self):
        self.props_index = {}
        if self.form.props:
            for prop in self.form.props:
                self.props_index[prop['name']] = prop['id']
                childs = prop.get('child', [])
                for child in childs:
                    self.props_index[f"{prop['name']}.{child['name']}"] = prop['id'], child['id']

    def create_commands_index_by_name(self):
        self.commands_index = {}
        if self.form.commands:
            for command in self.form.commands:
                self.commands_index[command['name']] = command['raw'][1]

    def decode_elements(self, raw_data):
        try:
            index = self.get_form_elem_index(raw_data)
            root_data = raw_data[1]
            self.create_prop_index_by_id()
            self.create_commands_index_by_id()
            # index_panel_count = index[1]
            # form_panels_count = int(root_data[index_panel_count])
            # if form_panels_count:
            #     self.command_panels = [root_data[index_panel_count + 1]]
            #     root_data[index_panel_count] = 'В отдельном файле'
            #     del root_data[index_panel_count + 1]

            index_root_element_count = index[0]
            form_items_count = int(root_data[index_root_element_count])
            if form_items_count:
                self.form.elements_tree = FormElement.decode_list(self, root_data, index_root_element_count)
                self.form.elements_data = dict(sorted(self.form.elements_data.items()))
            pass
        except Exception as err:
            raise ExtException(parent=err)

    def get_form_elem_index(self, raw_data):
        try:
            root_data = raw_data[1]
            index_command_panel_count = calc_offset([(18, 2), (3, 0)], root_data)
            command_panel_count = int(root_data[index_command_panel_count])
            index_root_elem_count = index_command_panel_count + command_panel_count + 1
            return index_root_elem_count, index_command_panel_count
        except Exception as err:
            raise ExtException(
                message='случай требующий анализа, предоставьте образец формы разработчикам',
                detail=f'{self.form.name}, {err}')

    def encode(self, src_dir, file_name, dest_dir, raw_data):
        try:
            # elements = helper.json_read(src_dir, f'{file_name}.elements{self.version}.json')
            elements = helper.json_read(src_dir, f'{file_name}.elem.json')
            try:
                self.form.elements_tree = elements.get('tree')
                self.form.elements_data = elements.get('data')
                self.form.props = elements.get('props')
                self.form.params = elements.get('params')
                self.form.commands = elements.get('commands')
            except (KeyError, TypeError):
                raise ExtException(message='Форма разобрана старым сборщиком (<0.16), разберите её новым сборщиком',
                                   action='Form9of.encode_elements')

            self.form.commands = FormCommands.encode_list(self, src_dir, file_name, None, raw_data)
            self.form.props = FormProps.encode_list(self, src_dir, file_name, None, raw_data)

            self.create_prop_index_by_name()
            self.create_commands_index_by_name()

            self.encode_elements(src_dir, file_name, dest_dir, raw_data)
            FormParams.encode_list(self, src_dir, file_name, None, raw_data)
        except Exception as err:
            raise ExtException(parent=err)

    def encode_elements(self, src_dir, file_name, dest_dir, raw_data):
        try:
            index = self.get_form_elem_index(raw_data)
            root_data = raw_data[1]

            # index_panel_count = index[1]
            # form_panels_count = int(root_data[index_panel_count])
            # if form_panels_count == 'В отдельном файле':
            #     self.command_panels = helper.json_read(src_dir, f'{file_name}.panels{version}.json')

            index_root_element_count = index[0]
            if root_data[index_root_element_count] == 'Дочерние элементы отдельно':
                # root_data[index_root_element_count] = str(len(self.elements_tree))
                FormElement.encode_list(self, self.form.elements_tree, root_data, index_root_element_count)
        except Exception as err:
            raise ExtException(parent=err)



from v8unpack.MetaDataObject.core.Simple import SimpleNameFolder
from .Form5 import Form5
from .Form9 import Form9


class Form(SimpleNameFolder):
    versions = {
        '5': Form5,
        '7': Form5,
        '9': Form9,
        '12': Form9,
        '13': Form9,
        '14': Form9,
    }
    # @classmethod
    # def read_header(cls, src_dir, src_file_name, data_id):
    #     return super().read_header(src_dir, f"{data_id['type']}{src_file_name}", data_id)

    @classmethod
    def get_form_root(cls, header_data):
        # return header_data[0][1][1]
        obj_version = header_data[0][1][0]
        if obj_version == '0':
            return header_data[0][1]
        elif obj_version == '1':
            return header_data[0][1][1]
        raise NotImplementedError()

    @classmethod
    def get_decode_header(cls, header_data):
        return cls.get_form_root(header_data)[1][1]

    @classmethod
    def get_version(cls, header_data, options):
        form_root = cls.get_form_root(header_data)
        form_version = form_root[1][0]
        return form_version


class Form1(Form):
    @classmethod
    def get_form_root(cls, header_data):
        return header_data[0][1]


class Form0(Form):
    @classmethod
    def get_form_root(cls, header_data):
        return header_data[0]



import os

from .. import helper
from ..ext_exception import ExtException
from ..MetaObject.Configuration import Configuration
from ..version import __version__


class ConfigurationExtension(Configuration):
    info = ['6', '8', '9']
    ext_code = {
        'con': '5',
        'app': '6',
        'ssn': '7'
    }
    _obj_info = {
        'a': 'a',
    }

    def decode(self, src_dir, dest_dir, *, version=None, **kwargs):
        self.header = {}
        root = helper.brace_file_read(src_dir, 'configinfo')
        self.header['configinfo'] = root
        self.header['v8unpack'] = __version__
        self.header['file_uuid'] = root[1][1]
        self.header['version'] = root[0][1]
        # self.header['versions'] = root[2]
        self.header['copyinfo'] = root[1]
        # self.header['copyinfo'][2] = b64decode(self.header['copyinfo'][2])[:-16].hex()
        self.header['header'] = helper.brace_file_read(src_dir, f'{self.header["file_uuid"]}')
        _form_header = self.get_decode_header(self.header['header'])
        helper.decode_header(self, _form_header, id_in_separate_file=False)
        product_version = self.header['header'][0][3][1][1][15]
        if version is None:
            self.header['compatibility_version'] = self.header['header'][0][3][1][1][43]
        else:
            self.header['compatibility_version'] = version
            self.header['header'][0][3][1][1][43] = version

        self.decode_code(src_dir)

        for i in self.info:  # хз что это
            file_name = f'{self.header["uuid"]}.{i}'
            if os.path.isdir(os.path.join(src_dir, file_name)):
                continue
            try:
                self.header[f'info{i}'] = helper.brace_file_read(src_dir, file_name)
            except FileNotFoundError:
                pass
        file_name = self.get_class_name_without_version()
        self._decode_info(src_dir, dest_dir, file_name)
        tasks = self.decode_includes(src_dir, dest_dir, '', self.header['header'])

        helper.txt_write(helper.str_decode(product_version), dest_dir, 'version.bin', encoding='utf-8')
        self.header['obj_version'] = self.obj_version
        helper.json_write(self.header, dest_dir, f'{self.get_class_name_without_version()}.json')
        self.write_decode_code(dest_dir, self.__class__.__name__)

        return tasks

    def encode(self, src_dir, dest_dir, *, file_name=None, include_index=None, file_list=None):
        try:
            self.header = helper.json_read(src_dir, f'{self.get_class_name_without_version()}.json')

            self.set_product_info(src_dir, file_name)

            # установка режима совместимости
            version = self.get_options('version')
            if version is not None:
                self.header['header'][0][3][1][1][43] = version
            prefix = self.get_options('prefix')
            if prefix is not None:
                self.header['header'][0][3][1][1][42] = helper.str_encode(prefix)

            # gui = self.get_options('gui')
            # if gui is not None:
            #     self.header['header'][0][3][1][1][38] = gui

            helper.check_version(__version__, self.header.get('v8unpack', ''))

            if include_index and self.get_options('auto_include'):
                self.fill_header_includes(include_index)

            # self.header['copyinfo'][2] = b64encode(bytes.fromhex(self.header['copyinfo'][2]+md5().digest().hex())).decode()
            root = [
                ["0", self.encode_version()],
                self.header['copyinfo'],
                # self.header['versions']
                ["____versions____"]
            ]
            self.encode_code(src_dir, self.__class__.__name__)
            self.write_encode_code(dest_dir)
            _file_name = self.get_class_name_without_version()
            self._encode_info(src_dir, _file_name, dest_dir)
            helper.brace_file_write(root, dest_dir, 'configinfo')
            helper.brace_file_write(self.header['header'], dest_dir, self.header["file_uuid"])
            for i in self.info:  # хз что это
                try:
                    helper.brace_file_write(self.header[f'info{i}'], dest_dir, f'{self.header["uuid"]}.{i}')
                except KeyError:
                    pass

            return None
        except Exception as err:
            raise ExtException(parent=err)





import os
import re

from .FormElements26.FormElements26 import FormElements26
from .FormElements27.FormElements27 import FormElements27
from .FormElements4.FormElements4 import FormElements4
from v8unpack.MetaDataObject.core.Simple import SimpleNameFolder
from v8unpack import helper
from v8unpack.ext_exception import ExtException
import json

OF = '0'  # обычные формы
UF = '1'  # управляемые формы


class FormCore(SimpleNameFolder):
    _obj_name = "Form"
    double_quotes = re.compile(r'("")')
    quotes = re.compile(r'(")')
    supported_form_versions = {
        # '0-2': FormElements2,
        # '0-23': FormElements26,# 801
        # '0-25': FormElements26,# 801
        '0-26': FormElements26,  # 801
        '0-27': FormElements27,  # OF
        '1': FormElements4,  # UF
        # '1-4-50': FormElements4,
        # '1-4-49': FormElements4,
        # '1-3-49': FormElements4
    }

    def __init__(self, *, meta_obj_class=None, obj_version=None, options=None):
        super().__init__(meta_obj_class=meta_obj_class, obj_version=obj_version, options=options)
        # self.form = []
        self.elements_tree = []
        self.elements_data = {}
        self.props_index = {}
        self.props = []
        self.params = []
        self.commands = []
        self._form = None

    @property
    def form(self):
        try:
            return self.header['form']
        except KeyError:
            self.header['form'] = []
        return self.header['form']

    @form.setter
    def form(self, value):
        self.header['form'] = value

    @classmethod
    def get_form_root(cls, header_data):
        return header_data[0][1][1]

    def get_decode_header(self, header_data):
        try:
            _form_root_getter = getattr(self.meta_obj_class, 'get_form_root')
        except Exception as err:
            _form_root_getter = self.get_form_root
        return _form_root_getter(header_data)[1][1]

    def decode_object(self, src_dir, file_name, dest_dir, dest_path, version, header_data):
        # self.decode_header(header_data)
        self.set_write_decode_mode(dest_dir, dest_path)
        self.decode_data(src_dir, file_name)

    def decode_form0(self, src_dir):
        file_name = f'{self.header["uuid"]}.0'
        _code_dir = os.path.join(src_dir, file_name)
        if os.path.exists(_code_dir):
            if os.path.isdir(_code_dir):
                self.decode_form0_from_dir(_code_dir)
            else:
                self.decode_form0_from_file(_code_dir, file_name)
                pass

    def decode_form1(self, src_dir):
        try:
            form = helper.brace_file_read(src_dir, f'{self.header["uuid"]}.1')
        except FileNotFoundError:
            return
        self.form.append(form)

    def decode_form0_from_dir(self, src_dir):
        self.form.append(helper.brace_file_read(src_dir, 'form'))
        self.code['obj'], encoding = self.read_raw_code(src_dir, 'module', uncomment_directive=True)
        self.header['code_encoding_obj'] = encoding  # можно безболезненно поменять на utf-8-sig
        self.header[f'code_info_obj'] = 1

    def decode_form0_from_file(self, src_dir, file_name=None):
        if file_name is None:
            file_name = f'{self.header["uuid"]}.0'
        try:
            form = helper.brace_file_read(src_dir, file_name)
        except FileNotFoundError:
            self.form.append([])
            return
        try:
            _code = helper.str_decode(self.getset_form_code(form, 'Код в отдельном файле', self.header))
            if _code is not None:
                _code = self.double_quotes.sub(r'"', _code)
                self.code['obj'] = _code
                self.header['code_info_obj'] = 'Код в отдельном файле'
        except Exception as err:
            raise ExtException(parent=err, detail=self.header['uuid'])
        self.form.append(form)

    @classmethod
    def getset_form_code(cls, form, new_value=None, header=None):
        err_detail = f'{header["uuid"]} {header["name"]} ' \
                     f'опытным путем подобрано, если у Вас код не достается' \
                     f'обновитесь до последней версии, и если не поможет создайте issue с дампом'
        len_form_0 = len(form[0])
        if len_form_0 > 2 and form[0][0] in ['4', '3', '2']:
            code = form[0][2]
            if not isinstance(code, str):
                raise ExtException(
                    message='Not supported forms',
                    detail=err_detail,
                    dump=form
                )
            form[0][2] = new_value
            if len_form_0 != 10:
                a = 1
            return code

        last_level = cls.get_last_level_array(form)
        if len_form_0 < 10 and (last_level[0] == '49' or last_level[0] == '4'):
            return ''

        if len(last_level) > 10 \
                and last_level[0] in ['22', '1'] \
                and last_level[-1] == '0' \
                and last_level[-2] == '0':
            code_index = -8
            code = last_level[code_index]
            if not isinstance(code, str):
                raise ExtException(
                    message='Not supported forms',
                    detail=err_detail,
                    dump=form
                )
            if new_value is not None:
                last_level[code_index] = new_value
            return code
        return ''

    def get_form_element_decoder(self):
        self.header['Версия элементов формы'] = ''
        if not self.form or not self.form[0]:
            return
        self.header['Версия элементов формы'] = self.header['Тип формы']
        try:
            return self.supported_form_versions[self.header['Версия элементов формы']]
        except KeyError:
            pass

        self.header['Версия элементов формы'] += f"-{self.form[0][0][0]}"
        try:
            return self.supported_form_versions[self.header['Версия элементов формы']]
        except KeyError:
            pass

        try:
            v2 = self.form[0][0][1][0]
            self.header['Версия элементов формы'] += f'-{v2}'
        except (KeyError, TypeError, AttributeError):
            pass

        try:
            return self.supported_form_versions[self.header['Версия элементов формы']]
        except KeyError:
            ver = self.header['Версия элементов формы']
            return

    def decode_includes(self, src_dir, dest_dir, dest_path, header_data):
        try:
            if self.header['Тип формы'] == OF:
                self.decode_form0(src_dir)
                self.decode_code(src_dir, uncomment_directive=True)
            else:
                self.decode_form0_from_file(src_dir)
            self.decode_form1(src_dir)

            form_element_decoder = self.get_form_element_decoder()
            if not form_element_decoder:
                return

            backup = json.dumps(self.form)
            try:
                form_element_decoder(self).decode(src_dir, dest_dir, dest_path, self.form[0][0])
            except Exception as err:
                self.form = json.loads(backup)
                pass  # todo если какие то елементы формы не разбираются, не прерываем
                # raise ExtException(parent=err, message='Ошибка при разборе формы')

            #
            #
            # if self.header.get('Тип формы') == OF:
            #     self.decode_old_elements()
            #     return
        except Exception as err:
            raise ExtException(parent=err)

    def write_decode_object(self, dest_dir, dest_path, file_name):
        dest_full_path = os.path.join(dest_dir, dest_path)
        id_data = {
            'uuid': self.header.pop('uuid'),
            # 'name': self.header.pop('name'),
        }
        self.header['obj_version'] = self.obj_version
        helper.json_write(id_data, dest_full_path, f'{file_name}.id.json')
        helper.json_write(self.header, dest_full_path, f'{file_name}.json')
        self.write_decode_code(dest_full_path, file_name)

        # helper.json_write(self.form, self.new_dest_dir, f"{typed_file_name}.form.json")
        # if self.elements_tree or self.props or self.params or self.commands:
        helper.json_write(
            dict(
                params=self.params,
                props=self.props,
                commands=self.commands,
                tree=self.elements_tree,
                data=self.elements_data,
            ),
            self.new_dest_dir, f"{file_name}.elem.json")
        return []

    def decode_data(self, src_dir, uuid):
        _header_obj = self.meta_obj_class.get_form_root(self.header['header'])
        # self.header['Версия формы'] = _header_obj[1][0]
        try:
            self.header['Тип формы'] = _header_obj[1][3]
        except IndexError:
            self.header['Тип формы'] = OF
        return _header_obj

    # def get_decode_header(self, header):
    #     return self.get_decode_obj_header(header)[1][1]
    #
    # def get_decode_obj_header(self, header_data):
    #     return header_data[0][1] if self.obj_version == '0' else header_data[0][1][1]

    def encode_object(self, src_dir, file_name, dest_dir):
        super(FormCore, self).encode_object(src_dir, file_name, dest_dir)
        # try:
        # version = self.options.get('version')
        # if version is None:
        # version = '803' if str(self.header['Тип формы']) == '1' else '802'
        # version = version[:3]

        # self.form = helper.json_read(src_dir, f'{file_name}.form.json')
        # self.form = helper.json_read(src_dir, f'{file_name}.form{version}.json')
        # except FileNotFoundError:
        #     self.form = None
        # raise ExtException(message='File not found', detail=f'{src_dir} {}')
        # self.form = self.encode_empty_form()
        self.encode_data()
        self.encode_nested_includes(src_dir, file_name, dest_dir, None)

    def encode_header(self):
        try:
            form_header = self.get_decode_header(self.header['header'])
            helper.encode_header(self, form_header)
            return self.header['header']
        except Exception as err:
            raise ExtException(parent=err)

    def encode_nested_includes(self, src_dir, file_name, dest_dir, parent_id):
        # if self.header.get('Тип формы') == OF:
        #     self.encode_old_elements(src_dir, file_name, dest_dir, parent_id)
        #     return

        if not self.form or not self.form[0]:
            return

        form_element_decoder = self.get_form_element_decoder()
        if not form_element_decoder:
            return

        # try:
        #     current_form_version = f'{self.form[0][0][0]}-{self.form[0][0][1][0]}'
        #     if current_form_version not in self.supported_form_versions:
        #         return
        # except:
        #     return
        try:
            form_element_decoder(self).encode(src_dir, file_name, dest_dir, self.form[0][0])
        except Exception as err:
            raise ExtException(parent=err, action='Сборка формы') from None

    def write_encode_object(self, dest_dir):
        try:
            if self.header['Тип формы'] == OF:
                self.write_old_encode_object(dest_dir)
            else:
                helper.brace_file_write(self.encode_header(), dest_dir, self.header["uuid"])
                self.file_list.append(self.header["uuid"])
                if self.form:
                    file_name = f'{self.header["uuid"]}.0'
                    helper.brace_file_write(self.form[0], dest_dir, file_name)
                    self.file_list.append(file_name)
            if self.form and len(self.form) > 1:
                file_name = f'{self.header["uuid"]}.1'
                helper.brace_file_write(self.form[1], dest_dir, file_name)
                self.file_list.append(file_name)
        except Exception as err:
            raise ExtException(parent=err,
                               detail=f'{self.__class__.__name__} {self.header["name"]} {self.header["uuid"]}')

    def encode_data(self):
        if self.header['Тип формы'] == OF:
            return
        try:
            _code = self.code.pop('obj', "")
            _code = self.quotes.sub(r'""', _code)
            form0 = self.form[0]
            self.getset_form_code(form0, helper.str_encode(_code), self.header)
        except IndexError:
            pass
        return self.form

    def encode_empty_form(self):
        raise NotImplemented()

    def encode_write(self, dest_dir):
        raise NotImplemented()

    def write_encode_code(self, dest_dir, *, comment_directive=False):
        pass

    def write_old_encode_object(self, dest_dir):
        helper.brace_file_write(self.encode_header(), dest_dir, self.header["uuid"])
        self.file_list.append(self.header["uuid"])
        if self.header.get('code_info_obj'):
            dir_name = f'{self.header["uuid"]}.0'
            _code_dir = os.path.join(dest_dir, dir_name)
            self.file_list.append(dir_name)
            helper.makedirs(_code_dir, exist_ok=True)
            helper.brace_file_write(self.form[0], _code_dir, 'form')
            _encoding = self.header.get('code_encoding_obj', 'utf-8-sig')
            self.write_raw_code(self.code.get('obj', ''), _code_dir, 'module', encoding=_encoding,
                                comment_directive=True)

    @classmethod
    def get_last_level_array(cls, data):
        while True:
            if isinstance(data[-1], list):
                return cls.get_last_level_array(data[-1])
            else:
                return data



from ..MetaDataObject.core.Container import Container


class AccumulationRegister(Container):
    ext_code = {
        'mgr': 2,  # модуль менеджера
        'obj': 1,  # Модуль набора записей
    }
    help_file_number = 0
    predefined_file_number = 3  # Агрегаты

    @classmethod
    def get_decode_header(cls, header_data):
        return header_data[0][1][13][1]
